@@ -44,8 +44,10 @@
 #include <osmo-bts/abis.h>
 #include <osmo-bts/bts.h>
 #include <osmo-bts/bts_model.h>
+#include <osmo-bts/bts_sm.h>
 #include <osmo-bts/dtx_dl_amr_fsm.h>
 #include <osmo-bts/pcuif_proto.h>
+#include <osmo-bts/pcu_if.h>
 #include <osmo-bts/rsl.h>
 #include <osmo-bts/oml.h>
 #include <osmo-bts/signal.h>
@@ -55,17 +57,14 @@
 #include <osmo-bts/nm_common_fsm.h>
 #include <osmo-bts/power_control.h>
 #include <osmo-bts/osmux.h>
+#include <osmo-bts/notification.h>
 
+#define MAX_TA_DEF	 63 /* default max Timing Advance value */
 #define MIN_QUAL_RACH	 50 /* minimum link quality (in centiBels) for Access Bursts */
 #define MIN_QUAL_NORM	 -5 /* minimum link quality (in centiBels) for Normal Bursts */
 
 static void bts_update_agch_max_queue_length(struct gsm_bts *bts);
 
-struct gsm_network bts_gsmnet = {
-	.bts_list = { &bts_gsmnet.bts_list, &bts_gsmnet.bts_list },
-	.num_bts = 0,
-};
-
 void *tall_bts_ctx;
 
 /* Table 3.1 TS 04.08: Values of parameter S */
@@ -98,6 +97,7 @@ static const struct rate_ctr_desc bts_ctr_desc[] = {
 	[BTS_CTR_RACH_RCVD] =		{"rach:rcvd", "Received RACH requests (Um)"},
 	[BTS_CTR_RACH_DROP] =		{"rach:drop", "Dropped RACH requests (Um)"},
 	[BTS_CTR_RACH_HO] =		{"rach:handover", "Received RACH requests (Handover)"},
+	[BTS_CTR_RACH_VGCS] =		{"rach:vgcs", "Received RACH requests (VGCS)"},
 	[BTS_CTR_RACH_CS] =		{"rach:cs", "Received RACH requests (CS/Abis)"},
 	[BTS_CTR_RACH_PS] =		{"rach:ps", "Received RACH requests (PS/PCU)"},
 
@@ -149,7 +149,6 @@ struct osmo_tdef abis_T_defs[] = {
 	{}
 };
 
-static const uint8_t bts_nse_timer_default[] = { 3, 3, 3, 3, 30, 3, 10 };
 static const uint8_t bts_cell_timer_default[] =
 				{ 3, 3, 3, 3, 3, 10, 3, 10, 3, 10, 3 };
 static const struct gprs_rlc_cfg rlc_cfg_default = {
@@ -207,15 +206,16 @@ const char *btsatttr2str(enum bts_attribute v)
 const struct value_string bts_impl_flag_desc[] = {
 	{ BTS_INTERNAL_FLAG_MS_PWR_CTRL_DSP,	"DSP/HW based MS Power Control Loop" },
 	{ BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB,	"Measurement and Payload data combined" },
+	{ BTS_INTERNAL_FLAG_NM_RCHANNEL_DEPENDS_RCARRIER, "OML RadioChannel MO depends on RadioCarrier MO" },
+	{ BTS_INTERNAL_FLAG_INTERF_MEAS,	"Uplink interference measurements" },
 	{ 0, NULL }
 };
 
+/* Ensure that all BTS_INTERNAL_FLAG_* entries are present in bts_impl_flag_desc[] */
+osmo_static_assert(ARRAY_SIZE(bts_impl_flag_desc) == _BTS_INTERNAL_FLAG_NUM + 1, _bts_impl_flag_desc);
+
 static int gsm_bts_talloc_destructor(struct gsm_bts *bts)
 {
-	if (bts->site_mgr.mo.fi) {
-		osmo_fsm_inst_free(bts->site_mgr.mo.fi);
-		bts->site_mgr.mo.fi = NULL;
-	}
 	if (bts->mo.fi) {
 		osmo_fsm_inst_free(bts->mo.fi);
 		bts->mo.fi = NULL;
@@ -226,19 +226,26 @@ static int gsm_bts_talloc_destructor(struct gsm_bts *bts)
 	}
 
 	bts_osmux_release(bts);
+
+	llist_del(&bts->list);
+	g_bts_sm->num_bts--;
 	return 0;
 }
 
-struct gsm_bts *gsm_bts_alloc(void *ctx, uint8_t bts_num)
+struct gsm_bts *gsm_bts_alloc(struct gsm_bts_sm *bts_sm, uint8_t bts_num)
 {
-	struct gsm_bts *bts = talloc_zero(ctx, struct gsm_bts);
-	int i;
+	struct gsm_bts *bts = talloc_zero(bts_sm, struct gsm_bts);
 
 	if (!bts)
 		return NULL;
 
 	talloc_set_destructor(bts, gsm_bts_talloc_destructor);
 
+	/* add to list of BTSs */
+	llist_add_tail(&bts->list, &bts_sm->bts_list);
+	g_bts_sm->num_bts++;
+
+	bts->site_mgr = bts_sm;
 	bts->nr = bts_num;
 	bts->num_trx = 0;
 	INIT_LLIST_HEAD(&bts->trx_list);
@@ -251,32 +258,19 @@ struct gsm_bts *gsm_bts_alloc(void *ctx, uint8_t bts_num)
 					       LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(bts->shutdown_fi, "bts%d", bts->nr);
 
-	bts->site_mgr.mo.fi = osmo_fsm_inst_alloc(&nm_bts_sm_fsm, bts, &bts->site_mgr,
-						  LOGL_INFO, "bts_sm");
-	gsm_mo_init(&bts->site_mgr.mo, bts, NM_OC_SITE_MANAGER,
-		    0xff, 0xff, 0xff);
-
+	/* NM BTS */
 	bts->mo.fi = osmo_fsm_inst_alloc(&nm_bts_fsm, bts, bts,
 					 LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(bts->mo.fi, "bts%d", bts->nr);
 	gsm_mo_init(&bts->mo, bts, NM_OC_BTS, bts->nr, 0xff, 0xff);
 
-	for (i = 0; i < ARRAY_SIZE(bts->gprs.nsvc); i++) {
-		bts->gprs.nsvc[i].bts = bts;
-		bts->gprs.nsvc[i].id = i;
-		gsm_mo_init(&bts->gprs.nsvc[i].mo, bts, NM_OC_GPRS_NSVC,
-				bts->nr, i, 0xff);
-	}
-	memcpy(&bts->gprs.nse.timer, bts_nse_timer_default,
-		sizeof(bts->gprs.nse.timer));
-	gsm_mo_init(&bts->gprs.nse.mo, bts, NM_OC_GPRS_NSE,
-			bts->nr, 0xff, 0xff);
-	memcpy(&bts->gprs.cell.timer, bts_cell_timer_default,
-		sizeof(bts->gprs.cell.timer));
-	gsm_mo_init(&bts->gprs.cell.mo, bts, NM_OC_GPRS_CELL,
-			bts->nr, 0xff, 0xff);
-	memcpy(&bts->gprs.cell.rlc_cfg, &rlc_cfg_default,
-		sizeof(bts->gprs.cell.rlc_cfg));
+	/* NM GPRS CELL */
+	bts->gprs.cell.mo.fi = osmo_fsm_inst_alloc(&nm_gprs_cell_fsm, bts, &bts->gprs.cell,
+						  LOGL_INFO, NULL);
+	osmo_fsm_inst_update_id_f(bts->gprs.cell.mo.fi, "gprs_cell%d-0", bts->nr);
+	gsm_mo_init(&bts->gprs.cell.mo, bts, NM_OC_GPRS_CELL, bts->nr, 0, 0xff);
+	memcpy(&bts->gprs.cell.rlc_cfg, &rlc_cfg_default, sizeof(bts->gprs.cell.rlc_cfg));
+	memcpy(&bts->gprs.cell.timer, bts_cell_timer_default, sizeof(bts->gprs.cell.timer));
 
 	/* create our primary TRX. It will be initialized during bts_init() */
 	bts->c0 = gsm_bts_trx_alloc(bts);
@@ -292,14 +286,14 @@ struct gsm_bts *gsm_bts_alloc(void *ctx, uint8_t bts_num)
 	return bts;
 }
 
-struct gsm_bts *gsm_bts_num(const struct gsm_network *net, int num)
+struct gsm_bts *gsm_bts_num(const struct gsm_bts_sm *bts_sm, int num)
 {
 	struct gsm_bts *bts;
 
-	if (num >= net->num_bts)
+	if (num >= bts_sm->num_bts)
 		return NULL;
 
-	llist_for_each_entry(bts, &net->bts_list, list) {
+	llist_for_each_entry(bts, &bts_sm->bts_list, list) {
 		if (bts->nr == num)
 			return bts;
 	}
@@ -315,19 +309,14 @@ int bts_init(struct gsm_bts *bts)
 	static int initialized = 0;
 	void *tall_rtp_ctx;
 
-	/* add to list of BTSs */
-	llist_add_tail(&bts->list, &bts_gsmnet.bts_list);
-
 	bts->band = GSM_BAND_1800;
 
 	INIT_LLIST_HEAD(&bts->agch_queue.queue);
 	bts->agch_queue.length = 0;
 
 	bts->ctrs = rate_ctr_group_alloc(bts, &bts_ctrg_desc, bts->nr);
-	if (!bts->ctrs) {
-		llist_del(&bts->list);
+	if (!bts->ctrs)
 		return -1;
-	}
 
 	/* enable management with default levels,
 	 * raise threshold to GSM_BTS_AGCH_QUEUE_THRESH_LEVEL_DISABLE to
@@ -345,21 +334,27 @@ int bts_init(struct gsm_bts *bts)
 	bts->rtp_port_range_next = bts->rtp_port_range_start;
 	bts->rtp_ip_dscp = -1;
 	bts->rtp_priority = -1;
+	bts->emit_hr_rfc5993 = true;
 
 	/* Default (fall-back) MS/BS Power control parameters */
 	power_ctrl_params_def_reset(&bts->bs_dpc_params, true);
 	power_ctrl_params_def_reset(&bts->ms_dpc_params, false);
 
 	/* configurable via OML */
+	bts->bsic = 0xff; /* invalid value, guarded by bsc_configured=false */
+	bts->bsic_configured = false;
 	bts->load.ccch.load_ind_period = 112;
 	bts->rtp_jitter_buf_ms = 100;
-	bts->max_ta = 63;
+	bts->max_ta = MAX_TA_DEF;
 	bts->ny1 = 4;
+	bts->ny2 = 4;
 	bts->t3105_ms = 300;
+	bts->t3115_ms = 300;
 	bts->min_qual_rach = MIN_QUAL_RACH;
 	bts->min_qual_norm = MIN_QUAL_NORM;
 	bts->max_ber10k_rach = 1707; /* 7 of 41 bits is Eb/N0 of 0 dB = 0.1707 */
 	bts->pcu.sock_path = talloc_strdup(bts, PCU_SOCK_DEFAULT);
+	bts->pcu.sock_wqueue_len_max = BTS_PCU_SOCK_WQUEUE_LEN_DEFAULT;
 	for (i = 0; i < ARRAY_SIZE(bts->t200_ms); i++)
 		bts->t200_ms[i] = oml_default_t200_ms[i];
 
@@ -367,15 +362,9 @@ int bts_init(struct gsm_bts *bts)
 	bts->radio_link_timeout.oml = 32;
 	bts->radio_link_timeout.current = bts->radio_link_timeout.oml;
 
-	/* Start with the site manager */
-	oml_mo_state_init(&bts->site_mgr.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
+	/* Start with the BTS */
 	oml_mo_state_init(&bts->mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
-
-	/* set BTS attr to dependency */
-	oml_mo_state_init(&bts->gprs.nse.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
-	oml_mo_state_init(&bts->gprs.cell.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
-	oml_mo_state_init(&bts->gprs.nsvc[0].mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
-	oml_mo_state_init(&bts->gprs.nsvc[1].mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
+	oml_mo_state_init(&bts->gprs.cell.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
 
 	/* allocate a talloc pool for ORTP to ensure it doesn't have to go back
 	 * to the libc malloc all the time */
@@ -384,10 +373,8 @@ int bts_init(struct gsm_bts *bts)
 
 	/* Osmux */
 	rc = bts_osmux_init(bts);
-	if (rc < 0) {
-		llist_del(&bts->list);
+	if (rc < 0)
 		return rc;
-	}
 
 	/* features implemented in 'common', available for all models,
 	 * order alphabetically */
@@ -398,15 +385,15 @@ int bts_init(struct gsm_bts *bts)
 	osmo_bts_set_feature(bts->features, BTS_FEAT_IPV6_NSVC);
 	osmo_bts_set_feature(bts->features, BTS_FEAT_PAGING_COORDINATION);
 
+	/* Maximum TA supported by the PHY (can be overridden by PHY specific code) */
+	bts->support.max_ta = MAX_TA_DEF;
+
 	rc = bts_model_init(bts);
-	if (rc < 0) {
-		llist_del(&bts->list);
+	if (rc < 0)
 		return rc;
-	}
 
 	/* TRX0 was allocated early during gsm_bts_alloc, not later through VTY */
 	bts_model_trx_init(bts->c0);
-	bts_gsmnet.num_bts++;
 
 	if (!initialized) {
 		osmo_signal_register_handler(SS_GLOBAL, bts_signal_cbfn, NULL);
@@ -423,6 +410,9 @@ int bts_init(struct gsm_bts *bts)
 	bts->smscb_queue_tgt_len = 2;
 	bts->smscb_queue_hyst = 2;
 
+	bts->asci.pos_nch = -ENOTSUP;
+	INIT_LLIST_HEAD(&bts->asci.notifications);
+
 	INIT_LLIST_HEAD(&bts->bsc_oml_hosts);
 
 	/* register DTX DL FSM */
@@ -440,33 +430,13 @@ int bts_init(struct gsm_bts *bts)
 /* main link is established, send status report */
 int bts_link_estab(struct gsm_bts *bts)
 {
-	int i, j;
+	LOGP(DOML, LOGL_INFO, "Main link established, sending NM Status\n");
 
-	LOGP(DOML, LOGL_INFO, "Main link established, sending NM Status.\n");
-
-	/* BTS SITE MGR becomes Offline (tx SW ACT Report), BTS is DEPENDENCY */
-	osmo_fsm_inst_dispatch(bts->site_mgr.mo.fi, NM_EV_SW_ACT, NULL);
-	osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_SW_ACT, NULL);
-
-	/* those should all be in DEPENDENCY */
-	oml_tx_state_changed(&bts->gprs.nse.mo);
-	oml_tx_state_changed(&bts->gprs.cell.mo);
-	oml_tx_state_changed(&bts->gprs.nsvc[0].mo);
-	oml_tx_state_changed(&bts->gprs.nsvc[1].mo);
-
-	/* All other objects start off-line until the BTS Model code says otherwise */
-	for (i = 0; i < bts->num_trx; i++) {
-		struct gsm_bts_trx *trx = gsm_bts_trx_num(bts, i);
-
-		oml_tx_state_changed(&trx->mo);
-		oml_tx_state_changed(&trx->bb_transc.mo);
-
-		for (j = 0; j < ARRAY_SIZE(trx->ts); j++) {
-			struct gsm_bts_trx_ts *ts = &trx->ts[j];
-
-			oml_tx_state_changed(&ts->mo);
-		}
-	}
+	/* Signal OML UP to BTS SITE MGR. It will automatically SW_ACT repoort
+	 * and become Disabled-Offline, then dispatch same event to its children
+	 * objects.
+	 */
+	osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_OML_UP, NULL);
 
 	return bts_model_oml_estab(bts);
 }
@@ -693,7 +663,7 @@ int bts_agch_enqueue(struct gsm_bts *bts, struct msgb *msg)
 	return 0;
 }
 
-struct msgb *bts_agch_dequeue(struct gsm_bts *bts)
+static struct msgb *bts_agch_dequeue(struct gsm_bts *bts)
 {
 	struct msgb *msg = msgb_dequeue(&bts->agch_queue.queue);
 	if (!msg)
@@ -759,12 +729,12 @@ static void compact_agch_queue(struct gsm_bts *bts)
 	return;
 }
 
-int bts_ccch_copy_msg(struct gsm_bts *bts, uint8_t *out_buf, struct gsm_time *gt,
-		      int is_ag_res)
+int bts_ccch_copy_msg(struct gsm_bts *bts, uint8_t *out_buf, struct gsm_time *gt, enum ccch_msgt ccch)
 {
 	struct msgb *msg = NULL;
 	int rc = 0;
 	int is_empty = 1;
+	const struct bts_agch_msg_cb *msg_cb;
 
 	/* Do queue house keeping.
 	 * This needs to be done every time a CCCH message is requested, since
@@ -773,26 +743,39 @@ int bts_ccch_copy_msg(struct gsm_bts *bts, uint8_t *out_buf, struct gsm_time *gt
 	 */
 	compact_agch_queue(bts);
 
-	/* Check for paging messages first if this is PCH */
-	if (!is_ag_res)
+	switch (ccch) {
+	case CCCH_MSGT_NCH:
+		/* Send NCH message, it has priority over AGCH and does not overlap with PCH. */
+		rc = bts_asci_notify_nch_gen_msg(bts, out_buf);
+		return rc;
+	case CCCH_MSGT_PCH:
+		/* Check whether the block may be overwritten by AGCH. */
 		rc = paging_gen_msg(bts->paging_state, out_buf, gt, &is_empty);
-
-	/* Check whether the block may be overwritten */
-	if (!is_empty)
-		return rc;
-
-	msg = bts_agch_dequeue(bts);
-	if (!msg)
-		return rc;
+		if (!is_empty)
+			return rc;
+		/* fall-through */
+	case CCCH_MSGT_AGCH:
+		/* If fallen here and the AGCH queue is empty, return empty PCH message. */
+		msg = bts_agch_dequeue(bts);
+		if (!msg)
+			return rc;
+		/* Continue to return AGCH message. */
+		break;
+	}
 
 	rate_ctr_inc2(bts->ctrs, BTS_CTR_AGCH_SENT);
 
+	/* Confirm sending of the AGCH message towards the PCU */
+	msg_cb = (struct bts_agch_msg_cb *) msg->cb;
+	if (msg_cb->confirm)
+		pcu_tx_data_cnf(msg_cb->msg_id, PCU_IF_SAPI_AGCH_2);
+
 	/* Copy AGCH message */
 	memcpy(out_buf, msgb_l3(msg), msgb_l3len(msg));
 	rc = msgb_l3len(msg);
 	msgb_free(msg);
 
-	if (is_ag_res)
+	if (ccch == CCCH_MSGT_AGCH)
 		bts->agch_queue.agch_msgs++;
 	else
 		bts->agch_queue.pch_msgs++;
@@ -820,30 +803,35 @@ struct gsm_time *get_time(struct gsm_bts *bts)
 	return &bts->gsm_time;
 }
 
-int bts_supports_cm(const struct gsm_bts *bts,
-		    const struct rsl_ie_chan_mode *cm)
+bool bts_supports_cm_speech(const struct gsm_bts *bts,
+			    const struct rsl_ie_chan_mode *cm)
 {
 	enum osmo_bts_features feature = _NUM_BTS_FEAT;
 
-	switch (cm->spd_ind) {
-	case RSL_CMOD_SPD_SIGN:
-		/* We assume that signalling support is mandatory,
-		 * there is no BTS_FEAT_* definition to check that. */
-		return 1;
-	case RSL_CMOD_SPD_SPEECH:
+	/* Stage 1: check support for the requested channel type */
+	switch (cm->chan_rt) {
+	case RSL_CMOD_CRT_TCH_GROUP_Bm:
+	case RSL_CMOD_CRT_TCH_GROUP_Lm:
+		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VGCS))
+			return false;
+		break;
+	case RSL_CMOD_CRT_TCH_BCAST_Bm:
+	case RSL_CMOD_CRT_TCH_BCAST_Lm:
+		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VBS))
+			return false;
+		break;
+	case RSL_CMOD_CRT_OSMO_TCH_VAMOS_Bm:
+	case RSL_CMOD_CRT_OSMO_TCH_VAMOS_Lm:
+		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VAMOS))
+			return false;
 		break;
-	case RSL_CMOD_SPD_DATA:
-	default:
-		return 0;
 	}
 
-	/* Before the requested pchan/cm combination can be checked, we need to
-	 * convert it to a feature identifier we can check */
+	/* Stage 2: check support for the requested codec */
 	switch (cm->chan_rt) {
 	case RSL_CMOD_CRT_OSMO_TCH_VAMOS_Bm:
-		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VAMOS))
-			return 0;
-		/* fall-through */
+	case RSL_CMOD_CRT_TCH_GROUP_Bm:
+	case RSL_CMOD_CRT_TCH_BCAST_Bm:
 	case RSL_CMOD_CRT_TCH_Bm:
 		switch (cm->chan_rate) {
 		case RSL_CMOD_SP_GSM1:
@@ -857,14 +845,13 @@ int bts_supports_cm(const struct gsm_bts *bts,
 			break;
 		default:
 			/* Invalid speech codec type => Not supported! */
-			return 0;
+			return false;
 		}
 		break;
 
 	case RSL_CMOD_CRT_OSMO_TCH_VAMOS_Lm:
-		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_VAMOS))
-			return 0;
-		/* fall-through */
+	case RSL_CMOD_CRT_TCH_GROUP_Lm:
+	case RSL_CMOD_CRT_TCH_BCAST_Lm:
 	case RSL_CMOD_CRT_TCH_Lm:
 		switch (cm->chan_rate) {
 		case RSL_CMOD_SP_GSM1:
@@ -875,7 +862,7 @@ int bts_supports_cm(const struct gsm_bts *bts,
 			break;
 		default:
 			/* Invalid speech codec type => Not supported! */
-			return 0;
+			return false;
 		}
 		break;
 
@@ -883,14 +870,59 @@ int bts_supports_cm(const struct gsm_bts *bts,
 		LOGP(DRSL, LOGL_ERROR,
 		     "Unhandled RSL channel type=0x%02x/rate=0x%02x\n",
 		     cm->chan_rt, cm->chan_rate);
-		return 0;
+		return false;
 	}
 
 	/* Check if the feature is supported by this BTS */
 	if (osmo_bts_has_feature(bts->features, feature))
-		return 1;
+		return true;
 
-	return 0;
+	return false;
+}
+
+static bool bts_supports_cm_data(const struct gsm_bts *bts,
+				 const struct rsl_ie_chan_mode *cm)
+{
+	switch (bts->variant) {
+	case BTS_OSMO_TRX:
+		switch (cm->chan_rate) {
+		/* TODO: RSL_CMOD_CSD_NT_14k5 */
+		/* TODO: RSL_CMOD_CSD_T_14k4 */
+		case RSL_CMOD_CSD_NT_12k0:
+		case RSL_CMOD_CSD_T_9k6:
+			if (cm->chan_rt != RSL_CMOD_CRT_TCH_Bm)
+				return false; /* invalid */
+			/* fall-through */
+		case RSL_CMOD_CSD_NT_6k0:
+		case RSL_CMOD_CSD_T_4k8:
+		case RSL_CMOD_CSD_T_2k4:
+		case RSL_CMOD_CSD_T_1k2:
+		case RSL_CMOD_CSD_T_600:
+		case RSL_CMOD_CSD_T_1200_75:
+			return true;
+		default:
+			return false;
+		}
+	default:
+		return 0;
+	}
+}
+
+bool bts_supports_cm(const struct gsm_bts *bts,
+		     const struct rsl_ie_chan_mode *cm)
+{
+	switch (cm->spd_ind) {
+	case RSL_CMOD_SPD_SIGN:
+		/* We assume that signalling support is mandatory,
+		 * there is no BTS_FEAT_* definition to check that. */
+		return true;
+	case RSL_CMOD_SPD_SPEECH:
+		return bts_supports_cm_speech(bts, cm);
+	case RSL_CMOD_SPD_DATA:
+		return bts_supports_cm_data(bts, cm);
+	default:
+		return false;
+	}
 }
 
 /* return the gsm_lchan for the CBCH (if it exists at all) */
