@@ -1810,9 +1810,20 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 	if (lchan->ho.active == HANDOVER_WAIT_FRAME)
 		handover_frame(lchan);
 
-	/* report first valid received frame to VGCS talker process */
-	if (rsl_chan_rt_is_asci(lchan->rsl_chan_rt) && lchan->asci.talker_active == VGCS_TALKER_WAIT_FRAME)
-		vgcs_talker_frame(lchan);
+	if (rsl_chan_rt_is_asci(lchan->rsl_chan_rt)) {
+		/* report first valid received frame to VGCS talker process */
+		if (lchan->asci.talker_active == VGCS_TALKER_WAIT_FRAME)
+			vgcs_talker_frame(lchan);
+		/* Do not forward any message that is received on the uplink to LAPD while
+		 * the uplink is not active. If the MS did not recognize (fast enough) that
+		 * the uplink is free, it may continue to transmit LAPD messages. A
+		 * response by LAPD to these messages is not desired and not required. If
+		 * LAPD would respond, it would cause stopping transmission of UPLINK FREE
+		 * messages. No MS could access the uplink anymore.
+		 */
+		if (lchan->asci.talker_active != VGCS_TALKER_ACTIVE)
+			return 0;
+	}
 
 	if (L1SAP_IS_LINK_SACCH(link_id))
 		le = &lchan->lapdm_ch.lapdm_acch;
@@ -1979,6 +1990,7 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 			send_ul_rtp_packet_data(lchan, fn, msg->data, msg->len);
 			break;
 		case RSL_CMOD_SPD_SIGN:
+			return 0; /* drop stale TCH.ind */
 		default: /* shall not happen */
 			OSMO_ASSERT(0);
 		}
@@ -2432,3 +2444,26 @@ int l1sap_chan_modify(struct gsm_bts_trx *trx, uint8_t chan_nr)
 
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_MODIFY, 0);
 }
+
+int l1sap_uplink_access(struct gsm_lchan *lchan, bool active)
+{
+	uint8_t chan_nr = gsm_lchan2chan_nr(lchan);
+	struct osmo_phsap_prim l1sap;
+
+	if (lchan->state != LCHAN_S_ACTIVE && !rsl_chan_rt_is_asci(lchan->rsl_chan_rt)) {
+		LOGPLCHAN(lchan, DL1C, LOGL_ERROR, "Channel %s is not an active ASCI type channel.\n",
+			  rsl_chan_nr_str(chan_nr));
+		return -EINVAL;
+	}
+
+	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "%s uplink access detection on channel %s\n",
+		  (active) ? "Activating" : "Deactivating", rsl_chan_nr_str(chan_nr));
+
+
+	memset(&l1sap, 0, sizeof(l1sap));
+	osmo_prim_init(&l1sap.oph, SAP_GSM_PH, PRIM_MPH_INFO, PRIM_OP_REQUEST, NULL);
+	l1sap.u.info.type = (active) ? PRIM_INFO_ACT_UL_ACC : PRIM_INFO_DEACT_UL_ACC;
+	l1sap.u.info.u.ulacc_req.chan_nr = chan_nr;
+
+	return l1sap_down(lchan->ts->trx, &l1sap);
+}
