@@ -33,6 +33,7 @@
 #include <osmo-bts/handover.h>
 #include <osmo-bts/l1sap.h>
 #include <osmo-bts/bts_model.h>
+#include <osmo-bts/asci.h>
 #include <errno.h>
 
 static const struct value_string lchan_s_names[] = {
@@ -216,8 +217,11 @@ void gsm_lchan_release(struct gsm_lchan *lchan, enum lchan_rel_act_kind rel_kind
 	if (lchan->state == LCHAN_S_NONE)
 		return;
 
-	/* release handover state */
+	/* release handover, listener and talker states */
 	handover_reset(lchan);
+	vgcs_talker_reset(lchan, false);
+	vgcs_listener_reset(lchan);
+	vgcs_uplink_free_reset(lchan);
 
 	lchan->rel_act_kind = rel_kind;
 
@@ -322,7 +326,7 @@ void lchan_set_state(struct gsm_lchan *lchan, enum gsm_lchan_state state)
 	}
 }
 
-/* See Table 10.5.25 of GSM04.08 */
+/* See 3GPP TS 44.018 Table 10.5.2.5.1 "Channel Description information element" */
 static uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
 			  uint8_t ts_nr, uint8_t lchan_nr)
 {
@@ -367,13 +371,13 @@ static uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
 		cbits = ABIS_RSL_CHAN_NR_CBITS_BCCH;
 		break;
 	case GSM_PCHAN_NONE:
-		LOGP(DRSL, LOGL_ERROR, "Physical channel %s not expected!\n",
-		     gsm_pchan_name(pchan));
-		cbits = 0x00;
+		LOGP(DRSL, LOGL_ERROR, "ts=%u,ss=%u Physical channel %s not expected!\n",
+		     ts_nr, lchan_nr, gsm_pchan_name(pchan));
+		cbits = 0x00; /* Reserved */
 		break;
 	default:
-		LOGP(DRSL, LOGL_ERROR, "Physical channel %s (0x%02x) not expected!\n",
-		     gsm_pchan_name(pchan), (int)pchan);
+		LOGP(DRSL, LOGL_ERROR, "ts=%u,ss=%u Physical channel %s (0x%02x) not expected!\n",
+		     ts_nr, lchan_nr, gsm_pchan_name(pchan), (int)pchan);
 		OSMO_ASSERT(0);
 		break;
 	}
