@@ -34,6 +34,7 @@
 
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/msgb.h>
+#include <osmocom/core/utils.h>
 #include <osmocom/gsm/protocol/gsm_12_21.h>
 #include <osmocom/gsm/abis_nm.h>
 #include <osmocom/gsm/tlv.h>
@@ -46,8 +47,10 @@
 #include <osmo-bts/oml.h>
 #include <osmo-bts/bts_model.h>
 #include <osmo-bts/bts.h>
+#include <osmo-bts/bts_sm.h>
 #include <osmo-bts/signal.h>
 #include <osmo-bts/phy_link.h>
+#include <osmo-bts/nm_common_fsm.h>
 
 #define LOGPFOH(ss, lvl, foh, fmt, args ...) LOGP(ss, lvl, "%s: " fmt, abis_nm_dump_foh(foh), ## args)
 #define DEBUGPFOH(ss, foh, fmt, args ...) LOGPFOH(ss, LOGL_DEBUG, foh, fmt, ## args)
@@ -131,6 +134,7 @@ int oml_send_msg(struct msgb *msg, int is_manuf)
 int oml_mo_send_msg(const struct gsm_abis_mo *mo, struct msgb *msg, uint8_t msg_type)
 {
 	struct abis_om_fom_hdr *foh;
+	struct gsm_bts *bts;
 
 	msg->l3h = msgb_push(msg, sizeof(*foh));
 	foh = (struct abis_om_fom_hdr *) msg->l3h;
@@ -138,17 +142,31 @@ int oml_mo_send_msg(const struct gsm_abis_mo *mo, struct msgb *msg, uint8_t msg_
 	foh->obj_class = mo->obj_class;
 	memcpy(&foh->obj_inst, &mo->obj_inst, sizeof(foh->obj_inst));
 
-	/* FIXME: This assumption may not always be correct */
-	msg->trx = mo->bts->c0;
+	/* Find and set OML TRX on msg: */
+	switch (mo->obj_class) {
+	case NM_OC_SITE_MANAGER:
+		/* Pick the first BTS: */
+		bts = gsm_bts_num(g_bts_sm, 0);
+		break;
+	default:
+		/* Other objects should have a valid BTS available: */
+		bts = gsm_bts_num(g_bts_sm, mo->obj_inst.bts_nr);
+	}
+	if (OSMO_UNLIKELY(!bts)) {
+		LOGPFOH(DOML, LOGL_NOTICE, foh,
+			"Sending FOM failed (no related BTS object found)\n");
+		return -EINVAL;
+	}
+	msg->trx = bts->c0;
 
 	DEBUGPFOH(DOML, foh, "Tx %s\n", get_value_string(abis_nm_msgtype_names, foh->msg_type));
 
 	return oml_send_msg(msg, 0);
 }
 
-static inline void add_bts_attrs(struct msgb *msg, const struct gsm_bts *bts)
+/* Put NM_ATT_SW_CONFIG as per 9.4.61 "SW Configuration" */
+static int add_att_sw_config(struct msgb *msg, const struct gsm_abis_mo *mo)
 {
-	uint16_t total_len = 0;
 	uint8_t *len;
 
 	/* Put NM_ATT_SW_CONFIG as per 9.4.61 "SW Configuration". */
@@ -157,17 +175,39 @@ static inline void add_bts_attrs(struct msgb *msg, const struct gsm_bts *bts)
 	/* We don't know the length yet, so we update it later. */
 	len = msgb_put(msg, 2);
 
-	total_len += abis_nm_put_sw_file(msg, "osmobts", PACKAGE_VERSION, true);
-	total_len += abis_nm_put_sw_file(msg, btsatttr2str(BTS_TYPE_VARIANT),
-					 btsvariant2str(bts->variant), true);
+	switch (mo->obj_class) {
+	case NM_OC_BTS:
+	{
+		const struct gsm_bts *bts = mo->bts;
 
-	if (strlen(bts->sub_model)) {
-		total_len += abis_nm_put_sw_file(msg, btsatttr2str(BTS_SUB_MODEL),
-						 bts->sub_model, true);
+		abis_nm_put_sw_file(msg, "osmobts", PACKAGE_VERSION, true);
+		abis_nm_put_sw_file(msg, btsatttr2str(BTS_TYPE_VARIANT),
+				    btsvariant2str(bts->variant), true);
+		if (strlen(bts->sub_model)) {
+			abis_nm_put_sw_file(msg, btsatttr2str(BTS_SUB_MODEL),
+					    bts->sub_model, true);
+		}
+		break;
+	}
+	case NM_OC_BASEB_TRANSC:
+	{
+		const struct gsm_bts_trx *trx = container_of(mo, struct gsm_bts_trx, bb_transc.mo);
+		const struct phy_instance *pinst = trx->pinst;
+		const char *phy_version;
+
+		phy_version = pinst && strlen(pinst->version) ? pinst->version : "Unknown";
+		abis_nm_put_sw_file(msg, btsatttr2str(TRX_PHY_VERSION), phy_version, true);
+		break;
+	}
+	default:
+		msgb_get(msg, 1 + 2); /* TL16 */
+		return -ENOTSUP;
 	}
 
 	/* Finally, update the length */
-	osmo_store16be(total_len, len);
+	osmo_store16be((uint16_t)(msg->tail - (len + 2)), len);
+
+	return 0;
 }
 
 /* Add BTS features as 3GPP TS 52.021 §9.4.30 Manufacturer Id */
@@ -177,97 +217,64 @@ static inline void add_bts_feat(struct msgb *msg, const struct gsm_bts *bts)
 	msgb_tl16v_put(msg, NM_ATT_MANUF_ID, len, bts->features->data);
 }
 
-static inline void add_trx_attr(struct msgb *msg, const struct gsm_bts_trx *trx)
+/* Add ip.access feature flags for the given MO */
+static int add_att_ipacc_features(struct msgb *msg, const struct gsm_abis_mo *mo)
 {
-	const struct phy_instance *pinst = trx_phy_instance(trx);
-	const char *phy_version;
-	uint16_t total_len;
+	const struct gsm_bts *bts = mo->bts;
+	const struct gsm_bts_trx *trx;
+	uint32_t val;
 	uint8_t *len;
 
-	/* Put NM_ATT_SW_CONFIG as per 9.4.61 "SW Configuration". */
-	msgb_v_put(msg, NM_ATT_SW_CONFIG);
+	msgb_v_put(msg, NM_ATT_IPACC_SUPP_FEATURES);
 
 	/* We don't know the length yet, so we update it later. */
 	len = msgb_put(msg, 2);
 
-	phy_version = pinst && strlen(pinst->version) ? pinst->version : "Unknown";
-	total_len = abis_nm_put_sw_file(msg, btsatttr2str(TRX_PHY_VERSION), phy_version, true);
+	switch (mo->obj_class) {
+	case NM_OC_BTS:
+		msgb_tv16_put(msg, NM_IPAC_EIE_MAX_TA, 1); /* TL16 */
+		msgb_put_u8(msg, (bts->support.max_ta >> 0) & 0xff);
+		break;
+	case NM_OC_RADIO_CARRIER:
+		trx = container_of(mo, struct gsm_bts_trx, mo);
+		msgb_tv16_put(msg, NM_IPAC_EIE_FREQ_BANDS, 1); /* TL16 */
+		msgb_put_u8(msg, (trx->support.freq_bands >> 0) & 0xff);
+		break;
+	case NM_OC_BASEB_TRANSC:
+		trx = container_of(mo, struct gsm_bts_trx, bb_transc.mo);
+		msgb_tv16_put(msg, NM_IPAC_EIE_CIPH_ALGOS, 1); /* TL16 */
+		msgb_put_u8(msg, bts->support.ciphers); /* LSB is A5/1 */
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_CHAN_TYPES, 2); /* TL16 */
+		msgb_put_u8(msg, (trx->support.chan_types >> 0) & 0xff);
+		msgb_put_u8(msg, (trx->support.chan_types >> 8) & 0xff);
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_CHAN_MODES, 3); /* TL16 */
+		msgb_put_u8(msg, (trx->support.chan_modes >>  0) & 0xff);
+		msgb_put_u8(msg, (trx->support.chan_modes >>  8) & 0xff);
+		msgb_put_u8(msg, (trx->support.chan_modes >> 16) & 0xff);
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_RTP_FEATURES, 1); /* TL16 */
+		val = NM_IPAC_F_RTP_FEAT_IR_64k;
+		msgb_put_u8(msg, (val >> 0) & 0xff);
+
+		msgb_tv16_put(msg, NM_IPAC_EIE_RSL_FEATURES, 1); /* TL16 */
+		val = NM_IPAC_F_RSL_FEAT_DYN_PDCH_ACT
+		    | NM_IPAC_F_RSL_FEAT_RTP_PT2;
+		msgb_put_u8(msg, (val >> 0) & 0xff);
+		break;
+	case NM_OC_GPRS_CELL:
+		msgb_tv16_put(msg, NM_IPAC_EIE_GPRS_CODING, 2); /* TL16 */
+		msgb_put_u8(msg, (bts->gprs.cell.support.gprs_codings >> 0) & 0xff);
+		msgb_put_u8(msg, (bts->gprs.cell.support.gprs_codings >> 8) & 0xff);
+		break;
+	default:
+		msgb_get(msg, 1 + 2); /* TL16 */
+		return -ENOTSUP;
+	}
 
 	/* Finally, update the length */
-	osmo_store16be(total_len, len);
-}
-
-/* Handle a list of attributes requested by the BSC, compose
- * TRX-specific Get Attribute Response IE as per 9.4.64. */
-static inline int handle_attrs_trx(struct msgb *out_msg, const struct gsm_bts_trx *trx,
-				   const uint8_t *attr, uint16_t attr_len)
-{
-	uint8_t num_unsupported = 0;
-	uint8_t *buf;
-	int i;
-
-	if (!trx) {
-		LOGP(DOML, LOGL_ERROR, "%s: O&M Get Attributes for unknown TRX\n", gsm_trx_name(trx));
-		return -NM_NACK_TRXNR_UNKN;
-	}
-
-	for (i = 0; i < attr_len; i++) {
-		switch (attr[i]) {
-		case NM_ATT_SW_CONFIG:
-			add_trx_attr(out_msg, trx);
-			break;
-		default:
-			LOGP(DOML, LOGL_ERROR, "%s: O&M Get Attributes [%u], %s is unsupported by TRX.\n",
-			     gsm_trx_name(trx), i, get_value_string(abis_nm_att_names, attr[i]));
-			/* Push this tag to the list of unsupported attributes */
-			buf = msgb_push(out_msg, 1);
-			*buf = attr[i];
-			num_unsupported++;
-		}
-	}
-
-	/* Push the amount of unsupported attributes */
-	buf = msgb_push(out_msg, 1);
-	*buf = num_unsupported;
-
-	return 0;
-}
-
-/* Handle a list of attributes requested by the BSC, compose
- * BTS-specific Get Attribute Response IE as per 9.4.64. */
-static inline int handle_attrs_bts(struct msgb *out_msg, const struct gsm_bts *bts,
-				   const uint8_t *attr, uint16_t attr_len)
-{
-	uint8_t num_unsupported = 0;
-	uint8_t *buf;
-	int i;
-
-	if (!bts) {
-		LOGP(DOML, LOGL_ERROR, "O&M Get Attributes for unknown BTS\n");
-		return -NM_NACK_BTSNR_UNKN;
-	}
-
-	for (i = 0; i < attr_len; i++) {
-		switch (attr[i]) {
-		case NM_ATT_SW_CONFIG:
-			add_bts_attrs(out_msg, bts);
-			break;
-		case NM_ATT_MANUF_ID:
-			add_bts_feat(out_msg, bts);
-			break;
-		default:
-			LOGP(DOML, LOGL_ERROR, "O&M Get Attributes [%u], %s is unsupported by BTS.\n", i,
-			     get_value_string(abis_nm_att_names, attr[i]));
-			/* Push this tag to the list of unsupported attributes */
-			buf = msgb_push(out_msg, 1);
-			*buf = attr[i];
-			num_unsupported++;
-		}
-	}
-
-	/* Push the amount of unsupported attributes */
-	buf = msgb_push(out_msg, 1);
-	*buf = num_unsupported;
+	osmo_store16be((uint16_t)(msg->tail - (len + 2)), len);
 
 	return 0;
 }
@@ -277,31 +284,52 @@ static int oml_tx_attr_resp(const struct gsm_abis_mo *mo,
 			    const uint8_t *attr, uint16_t attr_len)
 {
 	struct msgb *nmsg = oml_msgb_alloc();
-	const char *mo_name = gsm_abis_mo_name(mo);
+	unsigned int num_unsupported = 0;
 	int rc;
 
 	if (!nmsg)
 		return -NM_NACK_CANT_PERFORM;
 
-	switch (mo->obj_class) {
-	case NM_OC_BTS:
-		rc = handle_attrs_bts(nmsg, mo->bts, attr, attr_len);
-		break;
-	case NM_OC_BASEB_TRANSC:
-		rc = handle_attrs_trx(nmsg, gsm_bts_trx_num(mo->bts, mo->obj_inst.trx_nr), attr, attr_len);
-		break;
-	default:
-		LOGP(DOML, LOGL_ERROR, "%s: Unsupported MO class in Get Attribute Response\n",
-		     mo_name);
-		rc = -NM_NACK_OBJCLASS_NOTSUPP;
+	for (unsigned int i = 0; i < attr_len; i++) {
+		switch (attr[i]) {
+		case NM_ATT_OPER_STATE:
+			msgb_tv16_put(nmsg, attr[i], 1);
+			msgb_put_u8(nmsg, mo->nm_state.operational);
+			break;
+		case NM_ATT_ADM_STATE:
+			msgb_tv16_put(nmsg, attr[i], 1);
+			msgb_put_u8(nmsg, mo->nm_state.administrative);
+			break;
+		case NM_ATT_AVAIL_STATUS:
+			msgb_tv16_put(nmsg, attr[i], 1);
+			msgb_put_u8(nmsg, mo->nm_state.availability);
+			break;
+		case NM_ATT_SW_CONFIG:
+			if (add_att_sw_config(nmsg, mo) != 0)
+				goto unsupported;
+			break;
+		case NM_ATT_MANUF_ID:
+			if (mo->obj_class == NM_OC_BTS)
+				add_bts_feat(nmsg, mo->bts);
+			else
+				goto unsupported;
+			break;
+		case NM_ATT_IPACC_SUPP_FEATURES:
+			if (add_att_ipacc_features(nmsg, mo) != 0)
+				goto unsupported;
+			break;
+		default:
+unsupported:
+			LOGP(DOML, LOGL_ERROR, "%s: O&M Get Attributes [%u], %s is unsupported\n",
+			     gsm_abis_mo_name(mo), i, get_value_string(abis_nm_att_names, attr[i]));
+			/* Push this tag to the list of unsupported attributes */
+			msgb_push_u8(nmsg, attr[i]);
+			num_unsupported++;
+		}
 	}
 
-	if (rc < 0) {
-		LOGP(DOML, LOGL_ERROR, "%s: Tx Get Attribute Response FAILED with rc=%d\n",
-		     mo_name, rc);
-		msgb_free(nmsg);
-		return rc;
-	}
+	/* Push the amount of unsupported attributes */
+	msgb_push_u8(nmsg, num_unsupported);
 
 	/* Push Get Attribute Response Info TL (actually TV where V is L) */
 	msgb_tv16_push(nmsg, NM_ATT_GET_ARI, msgb_length(nmsg));
@@ -450,13 +478,13 @@ int oml_fom_ack_nack(struct msgb *msg, uint8_t cause)
 
 	/* alter message type */
 	if (cause) {
-		LOGPFOH(DOML, LOGL_NOTICE, foh, "Sending FOM NACK with cause %s.\n",
+		LOGPFOH(DOML, LOGL_NOTICE, foh, "Sending FOM NACK with cause %s\n",
 			abis_nm_nack_cause_name(cause));
 		foh->msg_type += 2; /* nack */
 		/* add cause */
 		msgb_tv_put(msg, NM_ATT_NACK_CAUSES, cause);
 	} else {
-		LOGPFOH(DOML, LOGL_DEBUG, foh, "Sending FOM ACK.\n");
+		LOGPFOH(DOML, LOGL_DEBUG, foh, "Sending FOM ACK\n");
 		foh->msg_type++; /* ack */
 	}
 
@@ -472,6 +500,15 @@ int oml_fom_ack_nack(struct msgb *msg, uint8_t cause)
 	return 1;
 }
 
+/* Copy msg before calling oml_fom_ack_nack(), which takes its ownership */
+int oml_fom_ack_nack_copy_msg(const struct msgb *old_msg, uint8_t cause)
+{
+	struct msgb *msg = msgb_copy(old_msg, "OML-ack_nack");
+	msg->trx = old_msg->trx;
+	oml_fom_ack_nack(msg, cause);
+	return 0;
+}
+
 /*
  * Formatted O&M messages
  */
@@ -507,6 +544,7 @@ static int oml_rx_get_attr(struct gsm_bts *bts, struct msgb *msg)
 	const struct gsm_abis_mo *mo;
 	struct tlv_parsed tp;
 	int rc;
+	enum abis_nm_nack_cause c;
 
 	if (!foh || !bts)
 		return -EINVAL;
@@ -514,10 +552,9 @@ static int oml_rx_get_attr(struct gsm_bts *bts, struct msgb *msg)
 	DEBUGPFOH(DOML, foh, "Rx GET ATTR\n");
 
 	/* Determine which OML object is addressed */
-	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
-	if (!mo) {
+	if ((mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL) {
 		LOGPFOH(DOML, LOGL_ERROR, foh, "Get Attributes for unknown Object Instance\n");
-		return oml_fom_ack_nack(msg, NM_NACK_OBJINST_UNKN);
+		return oml_fom_ack_nack(msg, c);
 	}
 
 	rc = oml_tlv_parse(&tp, foh->data, msgb_l3len(msg) - sizeof(*foh));
@@ -549,6 +586,7 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 	struct tlv_parsed tp, *tp_merged;
 	int rc, i;
 	const uint8_t *payload;
+	struct nm_fsm_ev_setattr_data ev_data;
 
 	DEBUGPFOH(DOML, foh, "Rx SET BTS ATTR\n");
 
@@ -566,7 +604,7 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 			oml_tx_failure_event_rep(&bts->mo, NM_SEVER_MAJOR, OSMO_EVT_WARN_SW_WARN,
 						 "Given ARFCN %u is not supported",
 						 arfcn);
-			LOGPFOH(DOML, LOGL_ERROR, foh, "Given ARFCN %u is not supported.\n", arfcn);
+			LOGPFOH(DOML, LOGL_ERROR, foh, "Given ARFCN %u is not supported\n", arfcn);
 			return oml_fom_ack_nack(msg, NM_NACK_FREQ_NOTAVAIL);
 		}
 	}
@@ -580,6 +618,7 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 
 	/* merge existing BTS attributes with new attributes */
 	tp_merged = osmo_tlvp_copy(bts->mo.nm_attr, bts);
+	talloc_set_name_const(tp_merged, "oml_bts_attr");
 	osmo_tlvp_merge(tp_merged, &tp);
 
 	/* Ask BTS driver to validate new merged attributes */
@@ -696,26 +735,64 @@ static int oml_rx_set_bts_attr(struct gsm_bts *bts, struct msgb *msg)
 	if (TLVP_PRES_LEN(&tp, NM_ATT_BTS_AIR_TIMER, 1)) {
 		uint8_t t3105 = *TLVP_VAL(&tp, NM_ATT_BTS_AIR_TIMER);
 		if (t3105 == 0) {
-			LOGPFOH(DOML, LOGL_NOTICE, foh, "T3105 must have a value != 0.\n");
+			LOGPFOH(DOML, LOGL_NOTICE, foh, "T3105 must have a value != 0\n");
 			return oml_fom_ack_nack(msg, NM_NACK_PARAM_RANGE);
 		}
 		bts->t3105_ms = t3105 * 10;
+		/* there are no OML IEs for T3115; let's use T3105 as HO detection is a similar procedure */
+		bts->t3115_ms = bts->t3105_ms;
 	}
 
 	/* 9.4.37 NY1 */
-	if (TLVP_PRES_LEN(&tp, NM_ATT_NY1, 1))
+	if (TLVP_PRES_LEN(&tp, NM_ATT_NY1, 1)) {
 		bts->ny1 = *TLVP_VAL(&tp, NM_ATT_NY1);
+		/* there are no OML IEs for NY2; let's use NY1 as HO detection is a similar procedure */
+		bts->ny2 = bts->ny1;
+	}
 
 	/* 9.4.8 BCCH ARFCN */
 	if (TLVP_PRES_LEN(&tp, NM_ATT_BCCH_ARFCN, 2))
 		bts->c0->arfcn = ntohs(tlvp_val16_unal(&tp, NM_ATT_BCCH_ARFCN));
 
 	/* 9.4.9 BSIC */
-	if (TLVP_PRES_LEN(&tp, NM_ATT_BSIC, 1))
-		bts->bsic = *TLVP_VAL(&tp, NM_ATT_BSIC);
+	if (TLVP_PRES_LEN(&tp, NM_ATT_BSIC, 1)) {
+		struct gsm_bts_trx *trx;
+		uint8_t bts_tsc;
 
-	/* call into BTS driver to apply new attributes to hardware */
-	return bts_model_apply_oml(bts, msg, tp_merged, NM_OC_BTS, bts);
+		bts->bsic = *TLVP_VAL(&tp, NM_ATT_BSIC);
+		bts->bsic_configured = true;
+		bts_tsc = BTS_TSC(bts);
+
+		/* Apply TSC update on each TS if required: */
+		llist_for_each_entry(trx, &bts->trx_list, list) { /* C0..n */
+			unsigned int tn;
+			for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+				struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+
+				/* First some config validation: */
+				if (ts->tsc_oml_configured &&
+				    ts->tsc_oml != bts_tsc &&
+				    !osmo_bts_has_feature(bts->features, BTS_FEAT_MULTI_TSC)) {
+					LOGPFOH(DOML, LOGL_ERROR, foh, "SET BTS ATTR: this BTS model does not "
+						"support TSC %u != BSIC-BCC %u (TSC %u)\n",
+						ts->tsc_oml, bts->bsic, bts_tsc);
+					return oml_fom_ack_nack(msg, NM_NACK_PARAM_RANGE);
+				}
+
+				/* Now update TS TSC if needed: */
+				gsm_ts_apply_configured_tsc(ts);
+			}
+		}
+	}
+
+	ev_data = (struct nm_fsm_ev_setattr_data){
+		.msg = msg,
+	};
+
+	rc = osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_RX_SETATTR, &ev_data);
+	if (rc < 0)
+		return oml_fom_ack_nack(msg, NM_NACK_CANT_PERFORM);
+	return rc;
 }
 
 /* 8.6.2 Set Radio Attributes has been received */
@@ -724,6 +801,7 @@ static int oml_rx_set_radio_attr(struct gsm_bts_trx *trx, struct msgb *msg)
 	struct abis_om_fom_hdr *foh = msgb_l3(msg);
 	struct tlv_parsed tp, *tp_merged;
 	int rc;
+	struct nm_fsm_ev_setattr_data ev_data;
 
 	DEBUGPFOH(DOML, foh, "Rx SET RADIO CARRIER ATTR\n");
 
@@ -735,8 +813,9 @@ static int oml_rx_set_radio_attr(struct gsm_bts_trx *trx, struct msgb *msg)
 		return oml_fom_ack_nack(msg, NM_NACK_INCORR_STRUCT);
 	}
 
-	/* merge existing BTS attributes with new attributes */
-	tp_merged = osmo_tlvp_copy(trx->mo.nm_attr, trx->bts);
+	/* merge existing TRX attributes with new attributes */
+	tp_merged = osmo_tlvp_copy(trx->mo.nm_attr, trx);
+	talloc_set_name_const(tp_merged, "oml_trx_attr");
 	osmo_tlvp_merge(tp_merged, &tp);
 
 	/* Ask BTS driver to validate new merged attributes */
@@ -746,7 +825,7 @@ static int oml_rx_set_radio_attr(struct gsm_bts_trx *trx, struct msgb *msg)
 		return oml_fom_ack_nack(msg, -rc);
 	}
 
-	/* Success: replace old BTS attributes with new */
+	/* Success: replace old TRX attributes with new */
 	talloc_free(trx->mo.nm_attr);
 	trx->mo.nm_attr = tp_merged;
 
@@ -795,14 +874,22 @@ static int oml_rx_set_radio_attr(struct gsm_bts_trx *trx, struct msgb *msg)
 		if (arfcn >= 1024) { /* 0 .. 1023 (1024 channels total) */
 			oml_tx_failure_event_rep(&trx->bts->mo, NM_SEVER_MAJOR, OSMO_EVT_WARN_SW_WARN,
 						 "Given ARFCN %u is unsupported", arfcn);
-			LOGPFOH(DOML, LOGL_NOTICE, foh, "Given ARFCN %u is unsupported.\n", arfcn);
+			LOGPFOH(DOML, LOGL_NOTICE, foh, "Given ARFCN %u is unsupported\n", arfcn);
 			return oml_fom_ack_nack(msg, NM_NACK_FREQ_NOTAVAIL);
 		}
 		trx->arfcn = arfcn;
 	}
 #endif
-	/* call into BTS driver to apply new attributes to hardware */
-	return bts_model_apply_oml(trx->bts, msg, tp_merged, NM_OC_RADIO_CARRIER, trx);
+
+	ev_data = (struct nm_fsm_ev_setattr_data){
+		.msg = msg,
+	};
+
+	rc = osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_RX_SETATTR, &ev_data);
+	if (rc < 0)
+		return oml_fom_ack_nack(msg, NM_NACK_CANT_PERFORM);
+	return rc;
+
 }
 
 static int handle_chan_comb(struct gsm_bts_trx_ts *ts, const uint8_t comb)
@@ -898,6 +985,7 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 	struct gsm_bts *bts = ts->trx->bts;
 	struct tlv_parsed tp, *tp_merged;
 	int rc, i;
+	struct nm_fsm_ev_setattr_data ev_data;
 
 	DEBUGPFOH(DOML, foh, "Rx SET CHAN ATTR\n");
 
@@ -911,7 +999,7 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 	/* Check frequency hopping parameters (HSN, MAIO, ARFCN list) */
 	if (TLVP_PRESENT(&tp, NM_ATT_HSN) || TLVP_PRESENT(&tp, NM_ATT_MAIO)) {
 		if (!osmo_bts_has_feature(bts->features, BTS_FEAT_HOPPING)) {
-			LOGPFOH(DOML, LOGL_ERROR, foh, "SET CHAN ATTR: Frequency hopping not supported.\n");
+			LOGPFOH(DOML, LOGL_ERROR, foh, "SET CHAN ATTR: Frequency hopping not supported\n");
 			return oml_fom_ack_nack(msg, NM_NACK_SPEC_IMPL_NOTSUPP);
 		}
 
@@ -945,12 +1033,13 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 
 	/* 9.4.52 Starting Time */
 	if (TLVP_PRESENT(&tp, NM_ATT_START_TIME)) {
-		LOGPFOH(DOML, LOGL_NOTICE, foh, "SET CHAN ATTR: Starting time not supported.\n");
+		LOGPFOH(DOML, LOGL_NOTICE, foh, "SET CHAN ATTR: Starting time not supported\n");
 		return oml_fom_ack_nack(msg, NM_NACK_SPEC_IMPL_NOTSUPP);
 	}
 
-	/* merge existing BTS attributes with new attributes */
-	tp_merged = osmo_tlvp_copy(ts->mo.nm_attr, bts);
+	/* merge existing CHAN attributes with new attributes */
+	tp_merged = osmo_tlvp_copy(ts->mo.nm_attr, ts->trx);
+	talloc_set_name_const(tp_merged, "oml_chan_attr");
 	osmo_tlvp_merge(tp_merged, &tp);
 
 	/* Call into BTS driver to check attribute values */
@@ -962,7 +1051,7 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 		return oml_fom_ack_nack(msg, -rc);
 	}
 
-	/* Success: replace old BTS attributes with new */
+	/* Success: replace old CHAN attributes with new */
 	talloc_free(ts->mo.nm_attr);
 	ts->mo.nm_attr = tp_merged;
 
@@ -981,27 +1070,38 @@ static int oml_rx_set_chan_attr(struct gsm_bts_trx_ts *ts, struct msgb *msg)
 
 	/* 9.4.60 TSC */
 	if (TLVP_PRES_LEN(&tp, NM_ATT_TSC, 1)) {
-		ts->tsc_oml = ts->tsc = *TLVP_VAL(&tp, NM_ATT_TSC);
-		if (ts->tsc != BTS_TSC(bts) &&
+		ts->tsc_oml = *TLVP_VAL(&tp, NM_ATT_TSC);
+		ts->tsc_oml_configured = true;
+	}
+
+	if (ts->tsc_oml_configured) {
+		if (bts->bsic_configured &&
+		    ts->tsc_oml != BTS_TSC(bts) &&
 		    !osmo_bts_has_feature(bts->features, BTS_FEAT_MULTI_TSC)) {
 			LOGPFOH(DOML, LOGL_ERROR, foh, "SET CHAN ATTR: this BTS model does not "
-				"support TSC %u != BSIC-BCC %u\n", ts->tsc, BTS_TSC(bts));
+				"support TSC %u != BSIC-BCC %u\n", ts->tsc_oml, BTS_TSC(bts));
 			talloc_free(tp_merged);
 			return oml_fom_ack_nack(msg, NM_NACK_PARAM_RANGE);
 		}
-	} else {
-		/* If there is no TSC specified, use the BCC */
-		ts->tsc_oml = ts->tsc = BTS_TSC(bts);
+		gsm_ts_apply_configured_tsc(ts);
 	}
-	LOGPFOH(DOML, LOGL_INFO, foh, "SET CHAN ATTR (TSC=%u pchan=%s",
-		ts->tsc, gsm_pchan_name(ts->pchan));
+
+	LOGPFOH(DOML, LOGL_INFO, foh, "SET CHAN ATTR (TSC=%d pchan=%s",
+		ts->tsc_oml_configured ? (int)ts->tsc_oml : -1,
+		gsm_pchan_name(ts->pchan));
 	if (ts->hopping.enabled)
 		LOGPC(DOML, LOGL_INFO, " hsn=%u maio=%u chan_num=%u",
 		      ts->hopping.hsn, ts->hopping.maio, ts->hopping.arfcn_num);
 	LOGPC(DOML, LOGL_INFO, ")\n");
 
-	/* call into BTS driver to apply new attributes to hardware */
-	return bts_model_apply_oml(bts, msg, tp_merged, NM_OC_CHANNEL, ts);
+	ev_data = (struct nm_fsm_ev_setattr_data){
+		.msg = msg,
+	};
+
+	rc = osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_RX_SETATTR, &ev_data);
+	if (rc < 0)
+		return oml_fom_ack_nack(msg, NM_NACK_CANT_PERFORM);
+	return rc;
 }
 
 /* 8.9.2 Opstart has been received */
@@ -1010,14 +1110,16 @@ static int oml_rx_opstart(struct gsm_bts *bts, struct msgb *msg)
 	struct abis_om_fom_hdr *foh = msgb_l3(msg);
 	struct gsm_abis_mo *mo;
 	void *obj;
+	int rc;
+	enum abis_nm_nack_cause c;
 
 	DEBUGPFOH(DOML, foh, "Rx OPSTART\n");
 
 	/* Step 1: Resolve MO by obj_class/obj_inst */
-	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
-	obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst);
-	if (!mo || !obj)
-		return oml_fom_ack_nack(msg, NM_NACK_OBJINST_UNKN);
+	if ((mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+		return oml_fom_ack_nack(msg, c);
+	if ((obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+		return oml_fom_ack_nack(msg, c);
 
 	/* Step 2: Do some global dependency/consistency checking */
 	if (mo->nm_state.operational == NM_OPSTATE_ENABLED) {
@@ -1025,8 +1127,13 @@ static int oml_rx_opstart(struct gsm_bts *bts, struct msgb *msg)
 		return oml_mo_opstart_ack(mo);
 	}
 
-	/* Step 3: Ask BTS driver to apply the opstart */
-	return bts_model_opstart(bts, mo, obj);
+	/* Make sure all NM objects already have an FSM implemented: */
+	OSMO_ASSERT(mo->fi);
+
+	rc = osmo_fsm_inst_dispatch(mo->fi, NM_EV_RX_OPSTART, NULL);
+	if (rc < 0)
+		return oml_fom_ack_nack(msg, NM_NACK_CANT_PERFORM);
+	return rc;
 }
 
 static int oml_rx_chg_adm_state(struct gsm_bts *bts, struct msgb *msg)
@@ -1037,6 +1144,7 @@ static int oml_rx_chg_adm_state(struct gsm_bts *bts, struct msgb *msg)
 	uint8_t adm_state;
 	void *obj;
 	int rc;
+	enum abis_nm_nack_cause c;
 
 	DEBUGPFOH(DOML, foh, "Rx CHG ADM STATE\n");
 
@@ -1054,10 +1162,10 @@ static int oml_rx_chg_adm_state(struct gsm_bts *bts, struct msgb *msg)
 	adm_state = *TLVP_VAL(&tp, NM_ATT_ADM_STATE);
 
 	/* Step 1: Resolve MO by obj_class/obj_inst */
-	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
-	obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst);
-	if (!mo || !obj)
-		return oml_fom_ack_nack(msg, NM_NACK_OBJINST_UNKN);
+	if ((mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+		return oml_fom_ack_nack(msg, c);
+	if ((obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+		return oml_fom_ack_nack(msg, c);
 
 	/* Step 2: Do some global dependency/consistency checking */
 	if (mo->nm_state.administrative == adm_state) {
@@ -1179,16 +1287,21 @@ static int down_fom(struct gsm_bts *bts, struct msgb *msg)
  * manufacturer related messages
  */
 
-static int oml_ipa_mo_set_attr_nse(void *obj, const struct tlv_parsed *tp)
+static int oml_ipa_mo_set_attr_nse(void *obj,
+				   const struct msgb *msg,
+				   const struct tlv_parsed *tp)
 {
-	struct gsm_bts *bts = container_of(obj, struct gsm_bts, gprs.nse);
+	struct gsm_gprs_nse *nse = obj;
+	struct gsm_bts *bts = gsm_gprs_nse_get_bts(nse);
+	struct nm_fsm_ev_setattr_data ev_data;
+	int rc;
 
 	if (TLVP_PRES_LEN(tp, NM_ATT_IPACC_NSEI, 2))
-		bts->gprs.nse.nsei =
+		nse->nsei =
 			ntohs(tlvp_val16_unal(tp, NM_ATT_IPACC_NSEI));
 
 	if (TLVP_PRES_LEN(tp, NM_ATT_IPACC_NS_CFG, 7)) {
-		memcpy(&bts->gprs.nse.timer,
+		memcpy(&nse->timer,
 		       TLVP_VAL(tp, NM_ATT_IPACC_NS_CFG), 7);
 	}
 
@@ -1197,15 +1310,27 @@ static int oml_ipa_mo_set_attr_nse(void *obj, const struct tlv_parsed *tp)
 		       TLVP_VAL(tp, NM_ATT_IPACC_BSSGP_CFG), 11);
 	}
 
+	ev_data = (struct nm_fsm_ev_setattr_data){
+		.msg = msg,
+	};
+	rc = osmo_fsm_inst_dispatch(nse->mo.fi, NM_EV_RX_SETATTR, &ev_data);
+	if (rc < 0)
+		return NM_NACK_CANT_PERFORM;
+
 	osmo_signal_dispatch(SS_GLOBAL, S_NEW_NSE_ATTR, bts);
 
 	return 0;
 }
 
-static int oml_ipa_mo_set_attr_cell(void *obj, const struct tlv_parsed *tp)
+static int oml_ipa_mo_set_attr_cell(void *obj,
+				    const struct msgb *msg,
+				    const struct tlv_parsed *tp)
 {
-	struct gsm_bts *bts = container_of(obj, struct gsm_bts, gprs.cell);
-	struct gprs_rlc_cfg *rlcc = &bts->gprs.cell.rlc_cfg;
+	struct gsm_gprs_cell *gprs_cell = obj;
+	struct gsm_bts *bts = gsm_gprs_cell_get_bts(gprs_cell);
+	struct gprs_rlc_cfg *rlcc = &gprs_cell->rlc_cfg;
+	struct nm_fsm_ev_setattr_data ev_data;
+	int rc;
 	const uint8_t *cur;
 	uint16_t _cur_s;
 
@@ -1219,8 +1344,7 @@ static int oml_ipa_mo_set_attr_cell(void *obj, const struct tlv_parsed *tp)
 	}
 
 	if (TLVP_PRES_LEN(tp, NM_ATT_IPACC_BVCI, 2))
-		bts->gprs.cell.bvci =
-			ntohs(tlvp_val16_unal(tp, NM_ATT_IPACC_BVCI));
+		gprs_cell->bvci = ntohs(tlvp_val16_unal(tp, NM_ATT_IPACC_BVCI));
 
 	if (TLVP_PRES_LEN(tp, NM_ATT_IPACC_RLC_CFG, 9)) {
 		cur = TLVP_VAL(tp, NM_ATT_IPACC_RLC_CFG);
@@ -1267,14 +1391,25 @@ static int oml_ipa_mo_set_attr_cell(void *obj, const struct tlv_parsed *tp)
 		rlcc->initial_mcs = *TLVP_VAL(tp, NM_ATT_IPACC_RLC_CFG_3);
 	}
 
+	ev_data = (struct nm_fsm_ev_setattr_data){
+		.msg = msg,
+	};
+	rc = osmo_fsm_inst_dispatch(gprs_cell->mo.fi, NM_EV_RX_SETATTR, &ev_data);
+	if (rc < 0)
+		return NM_NACK_CANT_PERFORM;
+
 	osmo_signal_dispatch(SS_GLOBAL, S_NEW_CELL_ATTR, bts);
 
 	return 0;
 }
 
-static int oml_ipa_mo_set_attr_nsvc(struct gsm_bts_gprs_nsvc *nsvc,
+static int oml_ipa_mo_set_attr_nsvc(struct gsm_gprs_nsvc *nsvc,
+				    const struct msgb *msg,
 				    const struct tlv_parsed *tp)
 {
+	struct nm_fsm_ev_setattr_data ev_data;
+	int rc;
+
 	if (TLVP_PRES_LEN(tp, NM_ATT_IPACC_NSVCI, 2))
 		nsvc->nsvci = ntohs(tlvp_val16_unal(tp, NM_ATT_IPACC_NSVCI));
 
@@ -1336,34 +1471,18 @@ static int oml_ipa_mo_set_attr_nsvc(struct gsm_bts_gprs_nsvc *nsvc,
 		}
 	}
 
+	ev_data = (struct nm_fsm_ev_setattr_data){
+		.msg = msg,
+	};
+	rc = osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_RX_SETATTR, &ev_data);
+	if (rc < 0)
+		return NM_NACK_CANT_PERFORM;
 
 	osmo_signal_dispatch(SS_GLOBAL, S_NEW_NSVC_ATTR, nsvc);
 
 	return 0;
 }
 
-static int oml_ipa_mo_set_attr(struct gsm_bts *bts, const struct gsm_abis_mo *mo,
-				void *obj, const struct tlv_parsed *tp)
-{
-	int rc;
-
-	switch (mo->obj_class) {
-	case NM_OC_GPRS_NSE:
-		rc = oml_ipa_mo_set_attr_nse(obj, tp);
-		break;
-	case NM_OC_GPRS_CELL:
-		rc = oml_ipa_mo_set_attr_cell(obj, tp);
-		break;
-	case NM_OC_GPRS_NSVC:
-		rc = oml_ipa_mo_set_attr_nsvc(obj, tp);
-		break;
-	default:
-		rc = NM_NACK_OBJINST_UNKN;
-	}
-
-	return rc;
-}
-
 static int oml_ipa_set_attr(struct gsm_bts *bts, struct msgb *msg)
 {
 	struct abis_om_fom_hdr *foh = msgb_l3(msg);
@@ -1371,48 +1490,66 @@ static int oml_ipa_set_attr(struct gsm_bts *bts, struct msgb *msg)
 	struct tlv_parsed tp, *tp_merged;
 	void *obj;
 	int rc;
+	enum abis_nm_nack_cause c;
 
 	DEBUGPFOH(DOML, foh, "Rx IPA SET ATTR\n");
 
 	rc = oml_tlv_parse(&tp, foh->data, msgb_l3len(msg) - sizeof(*foh));
 	if (rc < 0) {
-		mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
-		if (!mo)
-			return oml_fom_ack_nack(msg, NM_NACK_OBJINST_UNKN);
+		if ((mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+			return oml_fom_ack_nack(msg, c);
 		oml_tx_failure_event_rep(mo, NM_SEVER_MAJOR, OSMO_EVT_MAJ_UNSUP_ATTR,
 					 "New value for IPAC Set Attribute not supported\n");
 		return oml_fom_ack_nack(msg, NM_NACK_INCORR_STRUCT);
 	}
 
 	/* Resolve MO by obj_class/obj_inst */
-	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
-	obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst);
-	if (!mo || !obj)
-		return oml_fom_ack_nack(msg, NM_NACK_OBJINST_UNKN);
+	if ((mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+		return oml_fom_ack_nack(msg, c);
+	if ((obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst, &c)) == NULL)
+		return oml_fom_ack_nack(msg, c);
 
-	rc = oml_ipa_mo_set_attr(bts, mo, obj, &tp);
-	if (rc == 0) {
-		/* Success: replace old MO attributes with new */
-		/* merge existing MO attributes with new attributes */
-		tp_merged = osmo_tlvp_copy(mo->nm_attr, bts);
-		osmo_tlvp_merge(tp_merged, &tp);
-		talloc_free(mo->nm_attr);
-		mo->nm_attr = tp_merged;
+
+	switch (mo->obj_class) {
+	case NM_OC_GPRS_NSE:
+		rc =  oml_ipa_mo_set_attr_nse(obj, msg, &tp);
+		break;
+	case NM_OC_GPRS_CELL:
+		rc = oml_ipa_mo_set_attr_cell(obj, msg, &tp);
+		break;
+	case NM_OC_GPRS_NSVC:
+		rc = oml_ipa_mo_set_attr_nsvc(obj, msg, &tp);
+		break;
+	default:
+		rc = NM_NACK_OBJINST_UNKN;
 	}
 
-	return oml_fom_ack_nack(msg, rc);
+	if (rc != 0)
+		return oml_fom_ack_nack(msg, rc);
+
+	/* Success: replace old MO attributes with new */
+	/* merge existing MO attributes with new attributes */
+	tp_merged = osmo_tlvp_copy(mo->nm_attr, bts);
+	talloc_set_name_const(tp_merged, "oml_ipa_attr");
+	osmo_tlvp_merge(tp_merged, &tp);
+	talloc_free(mo->nm_attr);
+	mo->nm_attr = tp_merged;
+
+	return rc;
 }
 
-static int rx_oml_ipa_rsl_connect(struct gsm_bts_trx *trx, struct msgb *msg,
+static int rx_oml_ipa_rsl_connect(struct gsm_bts *bts, struct msgb *msg,
 				  const struct tlv_parsed *tp)
 {
-	struct e1inp_sign_link *oml_link = trx->bts->oml_link;
-	uint16_t port = IPA_TCP_PORT_RSL;
-	const char *trx_name = gsm_trx_name(trx);
+	struct e1inp_sign_link *oml_link = bts->oml_link;
+	const struct abis_om_fom_hdr *foh = msgb_l3(msg);
+	struct gsm_bts_trx *trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
+	struct gsm_bts_bb_trx *bb_transc;
+	const char *trx_name;
 	struct in_addr in;
-	int rc;
-
+	uint16_t port = IPA_TCP_PORT_RSL;
 	uint8_t stream_id = 0;
+	int rc;
 
 	if (TLVP_PRESENT(tp, NM_ATT_IPACC_DST_IP))
 		in.s_addr = tlvp_val32_unal(tp, NM_ATT_IPACC_DST_IP);
@@ -1421,20 +1558,29 @@ static int rx_oml_ipa_rsl_connect(struct gsm_bts_trx *trx, struct msgb *msg,
 
 	if (TLVP_PRESENT(tp, NM_ATT_IPACC_DST_IP_PORT))
 		port = ntohs(tlvp_val16_unal(tp, NM_ATT_IPACC_DST_IP_PORT));
+
 	if (TLVP_PRESENT(tp, NM_ATT_IPACC_STREAM_ID))
 		stream_id = *TLVP_VAL(tp, NM_ATT_IPACC_STREAM_ID);
 
-	LOGP(DOML, LOGL_INFO, "%s: Rx IPA RSL CONNECT IP=%s PORT=%u STREAM=0x%02x\n",
-	     trx_name, inet_ntoa(in), port, stream_id);
-
-	if (trx->bts->variant == BTS_OSMO_OMLDUMMY) {
-		rc = 0;
-		LOGP(DOML, LOGL_NOTICE, "%s: Not connecting RSL in OML-DUMMY!\n", trx_name);
-	} else {
-		trx->rsl_tei = stream_id;
-		rc = e1inp_ipa_bts_rsl_connect_n(oml_link->ts->line, inet_ntoa(in), port, trx->nr);
+	if (!trx) {
+		LOGP(DOML, LOGL_ERROR, "Rx IPA RSL CONNECT IP=%s PORT=%u STREAM=0x%02x for unknown TRX_NR=%u\n",
+		     inet_ntoa(in), port, stream_id, foh->obj_inst.trx_nr);
+		rc = NM_NACK_TRXNR_UNKN;
+		goto tx_ack_nack;
 	}
 
+	bb_transc = &trx->bb_transc;
+	osmo_sockaddr_str_from_in_addr(&bb_transc->rsl.rem_addrstr, &in, port);
+	bb_transc->rsl.tei = stream_id;
+
+	trx_name = gsm_trx_name(trx);
+	LOGP(DOML, LOGL_INFO, "%s: Rx IPA RSL CONNECT IP=%s PORT=%u STREAM=0x%02x\n",
+	     trx_name, bb_transc->rsl.rem_addrstr.ip, bb_transc->rsl.rem_addrstr.port,
+	     bb_transc->rsl.tei);
+
+	rc = 0;
+
+tx_ack_nack:
 	/* The ACK/NACK is expected to contain all IEs */
 	if (!TLVP_PRESENT(tp, NM_ATT_IPACC_DST_IP)) /* TV32 */
 		msgb_tv_fixed_put(msg, NM_ATT_IPACC_DST_IP, sizeof(in),
@@ -1444,12 +1590,7 @@ static int rx_oml_ipa_rsl_connect(struct gsm_bts_trx *trx, struct msgb *msg,
 	if (!TLVP_PRESENT(tp, NM_ATT_IPACC_STREAM_ID)) /* TV */
 		msgb_tv_put(msg, NM_ATT_IPACC_STREAM_ID, stream_id);
 
-	if (rc < 0) {
-		LOGP(DOML, LOGL_ERROR, "%s: Error in abis_open(RSL): %d\n", trx_name, rc);
-		return oml_fom_ack_nack(msg, NM_NACK_CANT_PERFORM);
-	}
-
-	return oml_fom_ack_nack(msg, 0);
+	return oml_fom_ack_nack(msg, rc);
 }
 
 static int down_mom(struct gsm_bts *bts, struct msgb *msg)
@@ -1498,8 +1639,7 @@ static int down_mom(struct gsm_bts *bts, struct msgb *msg)
 
 	switch (foh->msg_type) {
 	case NM_MT_IPACC_RSL_CONNECT:
-		trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
-		ret = rx_oml_ipa_rsl_connect(trx, msg, &tp);
+		ret = rx_oml_ipa_rsl_connect(bts, msg, &tp);
 		break;
 	case NM_MT_IPACC_SET_ATTR:
 		ret = oml_ipa_set_attr(bts, msg);
@@ -1603,122 +1743,112 @@ void gsm_mo_init(struct gsm_abis_mo *mo, struct gsm_bts *bts,
 	mo->nm_state.administrative = NM_STATE_LOCKED;
 }
 
-/* obtain the MO structure for a given object instance */
-struct gsm_abis_mo *
-gsm_objclass2mo(struct gsm_bts *bts, uint8_t obj_class,
-	    const struct abis_om_obj_inst *obj_inst)
+/* Obtain the MO structure for a given object instance
+ *  \param[out] c nack cause for reply in case of error. Ignored if NULL */
+struct gsm_abis_mo *gsm_objclass2mo(struct gsm_bts *bts, uint8_t obj_class,
+				    const struct abis_om_obj_inst *obj_inst,
+				    enum abis_nm_nack_cause *c)
 {
 	struct gsm_bts_trx *trx;
-	struct gsm_abis_mo *mo = NULL;
 
-	switch (obj_class) {
+	switch ((enum abis_nm_obj_class)obj_class) {
 	case NM_OC_BTS:
-		mo = &bts->mo;
-		break;
+		return &bts->mo;
 	case NM_OC_RADIO_CARRIER:
-		if (obj_inst->trx_nr >= bts->num_trx) {
-			return NULL;
-		}
-		trx = gsm_bts_trx_num(bts, obj_inst->trx_nr);
-		mo = &trx->mo;
-		break;
+		if (!(trx = gsm_bts_trx_num(bts, obj_inst->trx_nr)))
+			goto nm_nack_trxnr_unkn;
+		return &trx->mo;
 	case NM_OC_BASEB_TRANSC:
-		if (obj_inst->trx_nr >= bts->num_trx) {
-			return NULL;
-		}
-		trx = gsm_bts_trx_num(bts, obj_inst->trx_nr);
-		mo = &trx->bb_transc.mo;
-		break;
+		if (!(trx = gsm_bts_trx_num(bts, obj_inst->trx_nr)))
+			goto nm_nack_trxnr_unkn;
+		return &trx->bb_transc.mo;
 	case NM_OC_CHANNEL:
-		if (obj_inst->trx_nr >= bts->num_trx) {
-			return NULL;
-		}
-		trx = gsm_bts_trx_num(bts, obj_inst->trx_nr);
+		if (!(trx = gsm_bts_trx_num(bts, obj_inst->trx_nr)))
+			goto nm_nack_trxnr_unkn;
 		if (obj_inst->ts_nr >= TRX_NR_TS)
-			return NULL;
-		mo = &trx->ts[obj_inst->ts_nr].mo;
-		break;
+			goto nm_nack_objinst_unkn;
+		return &trx->ts[obj_inst->ts_nr].mo;
 	case NM_OC_SITE_MANAGER:
-		mo = &bts->site_mgr.mo;
-		break;
+		return &g_bts_sm->mo;
 	case NM_OC_GPRS_NSE:
-		mo = &bts->gprs.nse.mo;
-		break;
+		if (obj_inst->bts_nr > 0)
+			goto nm_nack_objinst_unkn;
+		return &g_bts_sm->gprs.nse.mo;
 	case NM_OC_GPRS_CELL:
-		mo = &bts->gprs.cell.mo;
-		break;
+		return &bts->gprs.cell.mo;
 	case NM_OC_GPRS_NSVC:
-		if (obj_inst->trx_nr >= ARRAY_SIZE(bts->gprs.nsvc))
-			return NULL;
-		mo = &bts->gprs.nsvc[obj_inst->trx_nr].mo;
-		break;
-	}
-	return mo;
-}
-
-/* obtain the gsm_nm_state data structure for a given object instance */
-struct gsm_nm_state *
-gsm_objclass2nmstate(struct gsm_bts *bts, uint8_t obj_class,
-		 const struct abis_om_obj_inst *obj_inst)
-{
-	struct gsm_abis_mo *mo;
-
-	mo = gsm_objclass2mo(bts, obj_class, obj_inst);
-	if (!mo)
+		if (obj_inst->bts_nr > 0)
+			goto nm_nack_objinst_unkn;
+		if (obj_inst->trx_nr >= ARRAY_SIZE(g_bts_sm->gprs.nse.nsvc))
+			goto nm_nack_objinst_unkn;
+		return &g_bts_sm->gprs.nse.nsvc[obj_inst->trx_nr].mo;
+	default:
+		if (c != NULL)
+			*c = NM_NACK_OBJCLASS_NOTSUPP;
 		return NULL;
+	}
 
-	return &mo->nm_state;
+nm_nack_trxnr_unkn:
+	if (c != NULL)
+		*c = NM_NACK_TRXNR_UNKN;
+	return NULL;
+nm_nack_objinst_unkn:
+	if (c != NULL)
+		*c = NM_NACK_OBJINST_UNKN;
+	return NULL;
 }
 
-/* obtain the in-memory data structure of a given object instance */
-void *
-gsm_objclass2obj(struct gsm_bts *bts, uint8_t obj_class,
-	     const struct abis_om_obj_inst *obj_inst)
+/* Obtain the in-memory data structure of a given object instance
+ *  \param[out] c nack cause for reply in case of error. Ignored if NULL */
+void *gsm_objclass2obj(struct gsm_bts *bts, uint8_t obj_class,
+		       const struct abis_om_obj_inst *obj_inst,
+		       enum abis_nm_nack_cause *c)
 {
 	struct gsm_bts_trx *trx;
-	void *obj = NULL;
 
-	switch (obj_class) {
+	switch ((enum abis_nm_obj_class)obj_class) {
 	case NM_OC_BTS:
-		obj = bts;
-		break;
+		return bts;
 	case NM_OC_RADIO_CARRIER:
-		if (obj_inst->trx_nr >= bts->num_trx) {
-			return NULL;
-		}
-		trx = gsm_bts_trx_num(bts, obj_inst->trx_nr);
-		obj = trx;
-		break;
+		if (!(trx = gsm_bts_trx_num(bts, obj_inst->trx_nr)))
+			goto nm_nack_trxnr_unkn;
+		return trx;
 	case NM_OC_BASEB_TRANSC:
-		if (obj_inst->trx_nr >= bts->num_trx) {
-			return NULL;
-		}
-		trx = gsm_bts_trx_num(bts, obj_inst->trx_nr);
-		obj = &trx->bb_transc;
-		break;
+		if (!(trx = gsm_bts_trx_num(bts, obj_inst->trx_nr)))
+			goto nm_nack_trxnr_unkn;
+		return &trx->bb_transc;
 	case NM_OC_CHANNEL:
-		if (obj_inst->trx_nr >= bts->num_trx) {
-			return NULL;
-		}
-		trx = gsm_bts_trx_num(bts, obj_inst->trx_nr);
+		if (!(trx = gsm_bts_trx_num(bts, obj_inst->trx_nr)))
+			goto nm_nack_trxnr_unkn;
 		if (obj_inst->ts_nr >= TRX_NR_TS)
-			return NULL;
-		obj = &trx->ts[obj_inst->ts_nr];
-		break;
+			goto nm_nack_objinst_unkn;
+		return &trx->ts[obj_inst->ts_nr];
 	case NM_OC_SITE_MANAGER:
-		obj = &bts->site_mgr;
-		break;
+		return g_bts_sm;
 	case NM_OC_GPRS_NSE:
-		obj = &bts->gprs.nse;
-		break;
+		if (obj_inst->bts_nr > 0)
+			goto nm_nack_objinst_unkn;
+		return &g_bts_sm->gprs.nse;
 	case NM_OC_GPRS_CELL:
-		obj = &bts->gprs.cell;
-		break;
+		return &bts->gprs.cell;
 	case NM_OC_GPRS_NSVC:
-		if (obj_inst->trx_nr >= ARRAY_SIZE(bts->gprs.nsvc))
-			return NULL;
-		obj = &bts->gprs.nsvc[obj_inst->trx_nr];
-		break;
+		if (obj_inst->bts_nr > 0)
+			goto nm_nack_objinst_unkn;
+		if (obj_inst->trx_nr >= ARRAY_SIZE(g_bts_sm->gprs.nse.nsvc))
+			goto nm_nack_objinst_unkn;
+		return &g_bts_sm->gprs.nse.nsvc[obj_inst->trx_nr];
+	default:
+		if (c != NULL)
+			*c = NM_NACK_OBJCLASS_NOTSUPP;
+		return NULL;
 	}
-	return obj;
+
+nm_nack_trxnr_unkn:
+	if (c != NULL)
+		*c = NM_NACK_TRXNR_UNKN;
+	return NULL;
+nm_nack_objinst_unkn:
+	if (c != NULL)
+		*c = NM_NACK_OBJINST_UNKN;
+	return NULL;
 }
