@@ -22,6 +22,7 @@
 #include "btsconfig.h"
 
 #include <inttypes.h>
+#include <limits.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -43,12 +44,14 @@
 #include <osmocom/core/sockaddr_str.h>
 #include <osmocom/trau/osmo_ortp.h>
 #include <osmocom/core/fsm.h>
+#include <osmocom/codec/codec.h>
 
 #include <osmo-bts/logging.h>
 #include <osmo-bts/gsm_data.h>
 #include <osmo-bts/phy_link.h>
 #include <osmo-bts/abis.h>
 #include <osmo-bts/bts.h>
+#include <osmo-bts/bts_sm.h>
 #include <osmo-bts/rsl.h>
 #include <osmo-bts/oml.h>
 #include <osmo-bts/signal.h>
@@ -68,6 +71,9 @@
 #define BTS_TRX_STR BTS_NR_STR TRX_NR_STR
 #define BTS_TRX_TS_STR BTS_TRX_STR TS_NR_STR
 #define BTS_TRX_TS_LCHAN_STR BTS_TRX_TS_STR LCHAN_NR_STR
+/* INT32_MAX, because osmo_wqueue_init takes int as an argument
+ * and INT_MAX can't be stringified as a decimal */
+#define BTS_CFG_PCU_SOCK_WQUEUE_LEN_MAX_MAX 2147483647
 
 #define X(x) (1 << x)
 
@@ -134,20 +140,6 @@ int bts_vty_go_parent(struct vty *vty)
 	return vty->node;
 }
 
-int bts_vty_is_config_node(struct vty *vty, int node)
-{
-	switch (node) {
-	case TRX_NODE:
-	case BTS_NODE:
-	case PHY_NODE:
-	case PHY_INST_NODE:
-	case OSMUX_NODE:
-		return 1;
-	default:
-		return 0;
-	}
-}
-
 static const char osmobts_copyright[] =
 	"Copyright (C) 2010-2011 by Harald Welte, Andreas Eversberg and On-Waves\r\n"
 	"Copyright (C) 2011-2022 by sysmocom - s.f.m.c. GmbH\r\n"
@@ -160,7 +152,6 @@ struct vty_app_info bts_vty_info = {
 	.version	= PACKAGE_VERSION,
 	.copyright	= osmobts_copyright,
 	.go_parent_cb	= bts_vty_go_parent,
-	.is_config_node	= bts_vty_is_config_node,
 	.usr_attr_desc	= {
 		[BTS_VTY_ATTR_NEW_LCHAN] = \
 			"This command applies for newly created lchans",
@@ -173,13 +164,6 @@ struct vty_app_info bts_vty_info = {
 	},
 };
 
-extern struct gsm_network bts_gsmnet;
-
-struct gsm_network *gsmnet_from_vty(struct vty *v)
-{
-	return &bts_gsmnet;
-}
-
 static struct cmd_node bts_node = {
 	BTS_NODE,
 	"%s(bts)# ",
@@ -418,9 +402,15 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 	vty_out(vty, " rtp port-range %u %u%s", bts->rtp_port_range_start,
 		bts->rtp_port_range_end, VTY_NEWLINE);
 	if (bts->rtp_ip_dscp != -1)
-		vty_out(vty, " rtp ip-dscp %i%s", bts->rtp_ip_dscp, VTY_NEWLINE);
+		vty_out(vty, " rtp ip-dscp %d%s", bts->rtp_ip_dscp, VTY_NEWLINE);
 	if (bts->rtp_priority != -1)
-		vty_out(vty, " rtp socket-priority %i%s", bts->rtp_priority, VTY_NEWLINE);
+		vty_out(vty, " rtp socket-priority %d%s", bts->rtp_priority, VTY_NEWLINE);
+	if (bts->rtp_nogaps_mode)
+		vty_out(vty, " rtp continuous-streaming%s", VTY_NEWLINE);
+	vty_out(vty, " %srtp internal-uplink-ecu%s",
+		bts->use_ul_ecu ? "" : "no ", VTY_NEWLINE);
+	vty_out(vty, " rtp hr-format %s%s",
+		bts->emit_hr_rfc5993 ? "rfc5993" : "ts101318", VTY_NEWLINE);
 	vty_out(vty, " paging queue-size %u%s", paging_get_queue_max(bts->paging_state),
 		VTY_NEWLINE);
 	vty_out(vty, " paging lifetime %u%s", paging_get_lifetime(bts->paging_state),
@@ -440,6 +430,10 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		vty_out(vty, " gsmtap-remote-host %s%s",
 			bts->gsmtap.remote_host,
 			VTY_NEWLINE);
+	if (bts->gsmtap.local_host != NULL)
+		vty_out(vty, " gsmtap-local-host %s%s",
+			bts->gsmtap.local_host,
+			VTY_NEWLINE);
 	for (i = 0; i < sizeof(uint32_t) * 8; i++) {
 		if (bts->gsmtap.sapi_mask & ((uint32_t) 1 << i)) {
 			sapi_buf = get_value_string_or_null(gsmtap_sapi_names, i);
@@ -461,6 +455,8 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		VTY_NEWLINE);
 	if (strcmp(bts->pcu.sock_path, PCU_SOCK_DEFAULT))
 		vty_out(vty, " pcu-socket %s%s", bts->pcu.sock_path, VTY_NEWLINE);
+	if (bts->pcu.sock_wqueue_len_max != BTS_CFG_PCU_SOCK_WQUEUE_LEN_MAX_MAX)
+		vty_out(vty, " pcu-socket-wqueue-length %u%s", bts->pcu.sock_wqueue_len_max, VTY_NEWLINE);
 	if (bts->supp_meas_toa256)
 		vty_out(vty, " supp-meas-info toa256%s", VTY_NEWLINE);
 	vty_out(vty, " smscb queue-max-length %d%s", bts->smscb_queue_max_len, VTY_NEWLINE);
@@ -476,7 +472,7 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		const struct phy_instance *pinst = trx_phy_instance(trx);
 		vty_out(vty, " trx %u%s", trx->nr, VTY_NEWLINE);
 
-		if (trx->power_params.user_gain_mdB)
+		if (tpp->user_gain_mdB)
 			vty_out(vty, "  user-gain %u mdB%s",
 				tpp->user_gain_mdB, VTY_NEWLINE);
 		vty_out(vty, "  power-ramp max-initial %d mdBm%s",
@@ -499,10 +495,9 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 
 static int config_write_bts(struct vty *vty)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts;
 
-	llist_for_each_entry(bts, &net->bts_list, list)
+	llist_for_each_entry(bts, &g_bts_sm->bts_list, list)
 		config_write_bts_single(vty, bts);
 
 	osmo_tdef_vty_groups_write(vty, "");
@@ -561,16 +556,15 @@ DEFUN_ATTR(cfg_bts,
 	   "BTS Number\n",
 	   CMD_ATTR_IMMEDIATE)
 {
-	struct gsm_network *gsmnet = gsmnet_from_vty(vty);
 	int bts_nr = atoi(argv[0]);
 	struct gsm_bts *bts;
 
-	if (bts_nr >= gsmnet->num_bts) {
+	if (bts_nr >= g_bts_sm->num_bts) {
 		vty_out(vty, "%% Unknown BTS number %u (num %u)%s",
-			bts_nr, gsmnet->num_bts, VTY_NEWLINE);
+			bts_nr, g_bts_sm->num_bts, VTY_NEWLINE);
 		return CMD_WARNING;
 	} else
-		bts = gsm_bts_num(gsmnet, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 
 	vty->index = bts;
 	vty->index_sub = &bts->description;
@@ -776,6 +770,62 @@ DEFUN_USRATTR(cfg_bts_rtp_priority,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_rtp_cont_stream,
+      cfg_bts_rtp_cont_stream_cmd,
+      "rtp continuous-streaming",
+      RTP_STR "Always emit an RTP packet every 20 ms\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->rtp_nogaps_mode = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_no_rtp_cont_stream,
+      cfg_bts_no_rtp_cont_stream_cmd,
+      "no rtp continuous-streaming",
+      NO_STR RTP_STR "Always emit an RTP packet every 20 ms\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->rtp_nogaps_mode = false;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_rtp_int_ul_ecu,
+      cfg_bts_rtp_int_ul_ecu_cmd,
+      "rtp internal-uplink-ecu",
+      RTP_STR "Apply a BTS-internal ECU to the uplink traffic frame stream\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->use_ul_ecu = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_no_rtp_int_ul_ecu,
+      cfg_bts_no_rtp_int_ul_ecu_cmd,
+      "no rtp internal-uplink-ecu",
+      NO_STR RTP_STR "Apply a BTS-internal ECU to the uplink traffic frame stream\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->use_ul_ecu = false;
+	return CMD_SUCCESS;
+}
+
+DEFUN_ATTR(cfg_bts_rtp_hr_format,
+	   cfg_bts_rtp_hr_format_cmd,
+	   "rtp hr-format (rfc5993|ts101318)",
+	   RTP_STR "HRv1 codec output format\n" "RFC 5993\n" "TS 101 318\n",
+	   CMD_ATTR_IMMEDIATE)
+{
+	struct gsm_bts *bts = vty->index;
+
+	bts->emit_hr_rfc5993 = !strcmp(argv[0], "rfc5993");
+	return CMD_SUCCESS;
+}
+
 #define PAG_STR "Paging related parameters\n"
 
 DEFUN_ATTR(cfg_bts_paging_queue_size,
@@ -964,7 +1014,7 @@ DEFUN_ATTR(cfg_bts_max_ber_rach, cfg_bts_max_ber_rach_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_bts_pcu_sock, cfg_bts_pcu_sock_cmd,
+DEFUN(cfg_bts_pcu_sock_path, cfg_bts_pcu_sock_path_cmd,
 	"pcu-socket PATH",
 	"Configure the PCU socket file/path name\n"
 	"UNIX socket path\n")
@@ -977,6 +1027,16 @@ DEFUN(cfg_bts_pcu_sock, cfg_bts_pcu_sock_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_pcu_sock_ql, cfg_bts_pcu_sock_ql_cmd,
+	"pcu-socket-wqueue-length <1-" OSMO_STRINGIFY_VAL(BTS_CFG_PCU_SOCK_WQUEUE_LEN_MAX_MAX) ">",
+	"Configure the PCU socket queue length\n"
+	"Queue length\n")
+{
+	struct gsm_bts *bts = vty->index;
+	bts->pcu.sock_wqueue_len_max = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
 DEFUN_ATTR(cfg_bts_supp_meas_toa256, cfg_bts_supp_meas_toa256_cmd,
 	   "supp-meas-info toa256",
 	   "Configure the RSL Supplementary Measurement Info\n"
@@ -1207,10 +1267,10 @@ static void bts_dump_vty_features(struct vty *vty, const struct gsm_bts *bts)
 
 	vty_out(vty, "  BTS model specific (internal) flags:%s", VTY_NEWLINE);
 
-	for (i = 0, no_features = true; i < sizeof(bts->flags) * 8; i++) {
-		if (bts_internal_flag_get(bts, (1 << i))) {
+	for (i = 0, no_features = true; i < _BTS_INTERNAL_FLAG_NUM; i++) {
+		if (bts_internal_flag_get(bts, i)) {
 			vty_out(vty, "    %03u ", i);
-			vty_out(vty, "%-40s%s", get_value_string(bts_impl_flag_desc, (1 << i)), VTY_NEWLINE);
+			vty_out(vty, "%-40s%s", get_value_string(bts_impl_flag_desc, i), VTY_NEWLINE);
 			no_features = false;
 		}
 	}
@@ -1249,7 +1309,7 @@ static void bts_dump_vty(struct vty *vty, const struct gsm_bts *bts)
 	vty_out(vty, "  NM State: ");
 	net_dump_nmstate(vty, &bts->mo.nm_state);
 	vty_out(vty, "  Site Mgr NM State: ");
-	net_dump_nmstate(vty, &bts->site_mgr.mo.nm_state);
+	net_dump_nmstate(vty, &g_bts_sm->mo.nm_state);
 	if (strnlen(bts->pcu_version, MAX_VERSION_LENGTH))
 		vty_out(vty, "  PCU version %s connected%s",
 			bts->pcu_version, VTY_NEWLINE);
@@ -1308,23 +1368,22 @@ DEFUN(show_bts, show_bts_cmd, "show bts [<0-255>]",
       SHOW_STR "Display information about a BTS\n"
       BTS_NR_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	int bts_nr;
 
 	if (argc != 0) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS '%s'%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts_dump_vty(vty, gsm_bts_num(net, bts_nr));
+		bts_dump_vty(vty, gsm_bts_num(g_bts_sm, bts_nr));
 		return CMD_SUCCESS;
 	}
 	/* print all BTS's */
-	for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++)
-		bts_dump_vty(vty, gsm_bts_num(net, bts_nr));
+	for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++)
+		bts_dump_vty(vty, gsm_bts_num(g_bts_sm, bts_nr));
 
 	return CMD_SUCCESS;
 }
@@ -1332,21 +1391,22 @@ DEFUN(show_bts, show_bts_cmd, "show bts [<0-255>]",
 static void gprs_dump_vty(struct vty *vty, const struct gsm_bts *bts)
 {
 	unsigned int i;
+	const struct gsm_gprs_nse *nse = &bts->site_mgr->gprs.nse;
 
 	/* GPRS parameters received from the BSC */
 	vty_out(vty, "BTS %u, RAC %u, NSEI %u, BVCI %u%s",
 		bts->nr, bts->gprs.rac,
-		bts->gprs.nse.nsei,
+		nse->nsei,
 		bts->gprs.cell.bvci,
 		VTY_NEWLINE);
 
 	vty_out(vty, "  Cell NM state: ");
 	net_dump_nmstate(vty, &bts->gprs.cell.mo.nm_state);
 	vty_out(vty, "  NSE NM state: ");
-	net_dump_nmstate(vty, &bts->gprs.nse.mo.nm_state);
+	net_dump_nmstate(vty, &nse->mo.nm_state);
 
-	for (i = 0; i < ARRAY_SIZE(bts->gprs.nsvc); i++) {
-		const struct gsm_bts_gprs_nsvc *nsvc = &bts->gprs.nsvc[i];
+	for (i = 0; i < ARRAY_SIZE(nse->nsvc); i++) {
+		const struct gsm_gprs_nsvc *nsvc = &nse->nsvc[i];
 
 		vty_out(vty, "  NSVC%u (NSVCI %u) NM state: ", i, nsvc->nsvci);
 		net_dump_nmstate(vty, &nsvc->mo.nm_state);
@@ -1381,10 +1441,9 @@ DEFUN(show_bts_gprs, show_bts_gprs_cmd,
       SHOW_STR "Display information about a BTS\n"
       BTS_NR_STR "GPRS/EGPRS configuration\n")
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts;
 
-	bts = gsm_bts_num(net, atoi(argv[0]));
+	bts = gsm_bts_num(g_bts_sm, atoi(argv[0]));
 	if (bts == NULL) {
 		vty_out(vty, "%% can't find BTS '%s'%s",
 			argv[0], VTY_NEWLINE);
@@ -1400,16 +1459,15 @@ DEFUN(test_send_failure_event_report, test_send_failure_event_report_cmd, "test
       "Various testing commands\n"
       "Send a test OML failure event report to the BSC\n" BTS_NR_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	int bts_nr = atoi(argv[0]);
 	const struct gsm_bts *bts;
 
-	if (bts_nr >= net->num_bts) {
+	if (bts_nr >= g_bts_sm->num_bts) {
 		vty_out(vty, "%% can't find BTS '%s'%s", argv[0], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
-	bts = gsm_bts_num(net, bts_nr);
+	bts = gsm_bts_num(g_bts_sm, bts_nr);
 	oml_tx_failure_event_rep(&bts->mo, NM_SEVER_MINOR, OSMO_EVT_WARN_SW_WARN, "test message sent from VTY");
 
 	return CMD_SUCCESS;
@@ -1421,9 +1479,8 @@ DEFUN_HIDDEN(radio_link_timeout, radio_link_timeout_cmd, "bts <0-0> radio-link-t
 		"Use infinite timeout (DANGEROUS: only use during testing!)\n"
 		"Number of lost SACCH blocks\n")
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	int bts_nr = atoi(argv[0]);
-	struct gsm_bts *bts = gsm_bts_num(net, bts_nr);
+	struct gsm_bts *bts = gsm_bts_num(g_bts_sm, bts_nr);
 
 	if (!bts) {
 		vty_out(vty, "%% can't find BTS '%s'%s", argv[0], VTY_NEWLINE);
@@ -1452,12 +1509,11 @@ DEFUN(bts_c0_power_red,
       "BCCH carrier power reduction operation\n"
       "Power reduction value (in dB, even numbers only)\n")
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	const int bts_nr = atoi(argv[0]);
 	const int red = atoi(argv[1]);
 	struct gsm_bts *bts;
 
-	bts = gsm_bts_num(net, atoi(argv[0]));
+	bts = gsm_bts_num(g_bts_sm, atoi(argv[0]));
 	if (bts == NULL) {
 		vty_out(vty, "%% No such BTS (%d)%s", bts_nr, VTY_NEWLINE);
 		return CMD_WARNING;
@@ -1588,10 +1644,10 @@ static void trx_dump_vty(struct vty *vty, const struct gsm_bts_trx *trx)
 
 	vty_out(vty, "  NM State: ");
 	net_dump_nmstate(vty, &trx->mo.nm_state);
-	vty_out(vty, "  RSL State: %s%s", trx->rsl_link? "connected" : "disconnected", VTY_NEWLINE);
+	vty_out(vty, "  RSL State: %s%s", trx->bb_transc.rsl.link ? "connected" : "disconnected", VTY_NEWLINE);
 	vty_out(vty, "  Baseband Transceiver NM State: ");
 	net_dump_nmstate(vty, &trx->bb_transc.mo.nm_state);
-	vty_out(vty, "  IPA stream ID: 0x%02x%s", trx->rsl_tei, VTY_NEWLINE);
+	vty_out(vty, "  IPA stream ID: 0x%02x%s", trx->bb_transc.rsl.tei, VTY_NEWLINE);
 }
 
 static inline void print_all_trx(struct vty *vty, const struct gsm_bts *bts)
@@ -1607,19 +1663,18 @@ DEFUN(show_trx,
 	SHOW_STR "Display information about a TRX\n"
 	BTS_TRX_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts = NULL;
 	int bts_nr, trx_nr;
 
 	if (argc >= 1) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS '%s'%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts = gsm_bts_num(net, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 	}
 	if (argc >= 2) {
 		trx_nr = atoi(argv[1]);
@@ -1637,8 +1692,8 @@ DEFUN(show_trx,
 		return CMD_SUCCESS;
 	}
 
-	for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++)
-		print_all_trx(vty, gsm_bts_num(net, bts_nr));
+	for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++)
+		print_all_trx(vty, gsm_bts_num(g_bts_sm, bts_nr));
 
 	return CMD_SUCCESS;
 }
@@ -1663,7 +1718,6 @@ DEFUN(show_ts,
 	SHOW_STR "Display information about a TS\n"
 	BTS_TRX_TS_STR)
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts = NULL;
 	const struct gsm_bts_trx *trx = NULL;
 	const struct gsm_bts_trx_ts *ts = NULL;
@@ -1672,12 +1726,12 @@ DEFUN(show_ts,
 	if (argc >= 1) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS '%s'%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts = gsm_bts_num(net, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 	}
 	if (argc >= 2) {
 		trx_nr = atoi(argv[1]);
@@ -1718,8 +1772,8 @@ DEFUN(show_ts,
 		}
 	} else {
 		/* Iterate over all BTS, TRX in each BTS, TS in each TRX */
-		for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++) {
-			bts = gsm_bts_num(net, bts_nr);
+		for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++) {
+			bts = gsm_bts_num(g_bts_sm, bts_nr);
 			for (trx_nr = 0; trx_nr < bts->num_trx; trx_nr++) {
 				trx = gsm_bts_trx_num(bts, trx_nr);
 				for (ts_nr = 0; ts_nr < TRX_NR_TS; ts_nr++) {
@@ -1937,6 +1991,19 @@ static void lchan_dump_full_vty(struct vty *vty, const struct gsm_lchan *lchan)
 	vty_out(vty, "  Channel Mode / Codec: %s%s",
 		gsm48_chan_mode_name(lchan->tch_mode),
 		VTY_NEWLINE);
+	if (lchan->tch_mode == GSM48_CMODE_SPEECH_AMR) {
+		const struct amr_multirate_conf *amr_mrc = &lchan->tch.amr_mr;
+		const struct gsm48_multi_rate_conf *mr_conf =
+			(const struct gsm48_multi_rate_conf *) amr_mrc->gsm48_ie;
+		vty_out(vty, "  AMR Multi-Rate Configuration: ICMI=%u, Start Mode=%u gsm48=%02x%02x%s",
+			mr_conf->icmi, mr_conf->smod, amr_mrc->gsm48_ie[0], amr_mrc->gsm48_ie[1], VTY_NEWLINE);
+		for (uint8_t i = 0; i < amr_mrc->num_modes; i++) {
+			const struct amr_mode *amode = &amr_mrc->mode[i];
+			vty_out(vty, "    AMR Mode %u (%s), threshold=%u, hysteresis=%u%s",
+				amode->mode, osmo_amr_type_name(amode->mode),
+				amode->threshold, amode->hysteresis, VTY_NEWLINE);
+		}
+	}
 
 	if (lchan->abis_ip.bound_ip) {
 		ia.s_addr = htonl(lchan->abis_ip.bound_ip);
@@ -2069,7 +2136,6 @@ static int dump_lchan_bts(const struct gsm_bts *bts, struct vty *vty,
 static int lchan_summary(struct vty *vty, int argc, const char **argv,
 			 void (*dump_cb)(struct vty *, const struct gsm_lchan *))
 {
-	const struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_bts *bts = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
 	const struct gsm_bts_trx *trx = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
 	const struct gsm_bts_trx_ts *ts = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
@@ -2079,12 +2145,12 @@ static int lchan_summary(struct vty *vty, int argc, const char **argv,
 	if (argc >= 1) {
 		/* use the BTS number that the user has specified */
 		bts_nr = atoi(argv[0]);
-		if (bts_nr >= net->num_bts) {
+		if (bts_nr >= g_bts_sm->num_bts) {
 			vty_out(vty, "%% can't find BTS %s%s", argv[0],
 				VTY_NEWLINE);
 			return CMD_WARNING;
 		}
-		bts = gsm_bts_num(net, bts_nr);
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 
 		if (argc == 1)
 			return dump_lchan_bts(bts, vty, dump_cb);
@@ -2125,8 +2191,8 @@ static int lchan_summary(struct vty *vty, int argc, const char **argv,
 		return CMD_SUCCESS;
 	}
 
-	for (bts_nr = 0; bts_nr < net->num_bts; bts_nr++) {
-		bts = gsm_bts_num(net, bts_nr);
+	for (bts_nr = 0; bts_nr < g_bts_sm->num_bts; bts_nr++) {
+		bts = gsm_bts_num(g_bts_sm, bts_nr);
 		dump_lchan_bts(bts, vty, dump_cb);
 	}
 
@@ -2152,8 +2218,7 @@ DEFUN(show_lchan_summary,
 	return lchan_summary(vty, argc, argv, lchan_dump_short_vty);
 }
 
-static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
-				       const char **argv, int idx)
+static struct gsm_lchan *resolve_lchan(const char **argv, int idx)
 {
 	int bts_nr = atoi(argv[idx+0]);
 	int trx_nr = atoi(argv[idx+1]);
@@ -2164,7 +2229,7 @@ static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
 	struct gsm_bts_trx *trx;
 	struct gsm_bts_trx_ts *ts;
 
-	bts = gsm_bts_num(net, bts_nr);
+	bts = gsm_bts_num(g_bts_sm, bts_nr);
 	if (!bts)
 		return NULL;
 
@@ -2217,6 +2282,22 @@ DEFUN(cfg_bts_gsmtap_remote_host,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_gsmtap_local_host,
+      cfg_bts_gsmtap_local_host_cmd,
+      "gsmtap-local-host HOSTNAME",
+      "Enable local bind for GSMTAP Um logging (see also 'gsmtap-sapi')\n"
+      "Local IP address or hostname\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	osmo_talloc_replace_string(bts, &bts->gsmtap.local_host, argv[0]);
+
+	if (vty->type != VTY_FILE)
+		vty_out(vty, "%% This command requires restart%s", VTY_NEWLINE);
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_bts_no_gsmtap_remote_host,
       cfg_bts_no_gsmtap_remote_host_cmd,
       "no gsmtap-remote-host",
@@ -2234,6 +2315,24 @@ DEFUN(cfg_bts_no_gsmtap_remote_host,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_bts_no_gsmtap_local_host,
+      cfg_bts_no_gsmtap_local_host_cmd,
+      "no gsmtap-local-host",
+      NO_STR "Disable local bind for GSMTAP Um logging\n")
+{
+	struct gsm_bts *bts = vty->index;
+
+	if (bts->gsmtap.local_host != NULL)
+		talloc_free(bts->gsmtap.local_host);
+
+	bts->gsmtap.local_host = NULL;
+
+	if (vty->type != VTY_FILE)
+		vty_out(vty, "%% This command requires restart%s", VTY_NEWLINE);
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_bts_gsmtap_sapi_all, cfg_bts_gsmtap_sapi_all_cmd,
 	"gsmtap-sapi (enable-all|disable-all)",
 	"Enable/disable sending of UL/DL messages over GSMTAP\n"
@@ -2395,11 +2494,10 @@ DEFUN(bts_t_t_l_jitter_buf,
 	BTS_T_T_L_STR "RTP settings\n"
 	"Jitter buffer\n" "Size of jitter buffer in (ms)\n")
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct gsm_lchan *lchan;
 	int jitbuf_ms = atoi(argv[4]), rc;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2428,10 +2526,9 @@ DEFUN_ATTR(bts_t_t_l_loopback,
 	   BTS_T_T_L_STR "Set loopback\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2447,10 +2544,9 @@ DEFUN_ATTR(no_bts_t_t_l_loopback,
 	   NO_STR BTS_T_T_L_STR "Set loopback\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2473,13 +2569,12 @@ DEFUN_ATTR(bts_t_t_l_power_ctrl_mode,
 	   "Enable the power control loop\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	const struct gsm_power_ctrl_params *params;
 	struct lchan_power_ctrl_state *state;
 	const char **args = argv + 4;
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2510,12 +2605,11 @@ DEFUN_ATTR(bts_t_t_l_power_ctrl_current_max,
 	   "BS power reduction (in dB) or MS power level\n",
 	   CMD_ATTR_HIDDEN)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
 	struct lchan_power_ctrl_state *state;
 	const char **args = argv + 4;
 	struct gsm_lchan *lchan;
 
-	lchan = resolve_lchan(net, argv, 0);
+	lchan = resolve_lchan(argv, 0);
 	if (!lchan) {
 		vty_out(vty, "%% Could not resolve logical channel%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -2629,6 +2723,11 @@ int bts_vty_init(void *ctx)
 	install_element(BTS_NODE, &cfg_bts_rtp_port_range_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_ip_dscp_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_priority_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_cont_stream_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_rtp_cont_stream_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_int_ul_ecu_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_rtp_int_ul_ecu_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_hr_format_cmd);
 	install_element(BTS_NODE, &cfg_bts_band_cmd);
 	install_element(BTS_NODE, &cfg_description_cmd);
 	install_element(BTS_NODE, &cfg_no_description_cmd);
@@ -2643,7 +2742,8 @@ int bts_vty_init(void *ctx)
 	install_element(BTS_NODE, &cfg_bts_min_qual_rach_cmd);
 	install_element(BTS_NODE, &cfg_bts_min_qual_norm_cmd);
 	install_element(BTS_NODE, &cfg_bts_max_ber_rach_cmd);
-	install_element(BTS_NODE, &cfg_bts_pcu_sock_cmd);
+	install_element(BTS_NODE, &cfg_bts_pcu_sock_path_cmd);
+	install_element(BTS_NODE, &cfg_bts_pcu_sock_ql_cmd);
 	install_element(BTS_NODE, &cfg_bts_supp_meas_toa256_cmd);
 	install_element(BTS_NODE, &cfg_bts_no_supp_meas_toa256_cmd);
 	install_element(BTS_NODE, &cfg_bts_smscb_max_qlen_cmd);
@@ -2652,6 +2752,8 @@ int bts_vty_init(void *ctx)
 
 	install_element(BTS_NODE, &cfg_bts_gsmtap_remote_host_cmd);
 	install_element(BTS_NODE, &cfg_bts_no_gsmtap_remote_host_cmd);
+	install_element(BTS_NODE, &cfg_bts_gsmtap_local_host_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_gsmtap_local_host_cmd);
 	install_element(BTS_NODE, &cfg_bts_gsmtap_sapi_all_cmd);
 	install_element(BTS_NODE, &cfg_bts_gsmtap_sapi_cmd);
 	install_element(BTS_NODE, &cfg_bts_no_gsmtap_sapi_cmd);
